# -*- coding: utf-8 -*-
"""
Copyright (C) 2014-2015 Benjamin Eltzner

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
or see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import numpy.random as arandom
import numpy.linalg as la
import scipy.optimize as opt
import scipy.stats as stat
import math, sys, warnings, scipy
from math import sqrt, acos, log, exp, sin
from scipy.integrate import quad

################################################################################
################################   Constants   #################################
################################################################################

PI = np.pi
DEG = math.degrees(1)
EPS = 1e-9
OLD_SCIPY = [int(x) for x in scipy.__version__.split('.')]
OLD_SCIPY = (OLD_SCIPY[0] <= 0 and OLD_SCIPY[1] < 15)

################################################################################
###################################   PNS   ####################################
################################################################################

def fit (f, d, initial, verbose, small=True):
    warnings.filterwarnings('error')
    if verbose:
        print('Starting fit... ',)
        sys.stdout.flush()
    try:
        initial, exit_code = opt.leastsq(f, initial)
    except:
        if verbose: print('Exception in fit! Setting invalid exit code.')
        exit_code = 6
    fails = 0
    counter = 0
    while exit_code > 1 or (small and abs(initial[-1]) > 1):
        if verbose:
            print('failed "get_sphere" with exit code:', exit_code, '\nRestarting... ')
        fails += 1
        if fails > 3 or exit_code == 6:
            if verbose:
                print('with new start value... ')
            initial = 2 * arandom.rand(d + 1) - 1
            fails = 0
        try:
            initial, exit_code = opt.leastsq(f, initial)
        except:
            if verbose: print('Exception in fit! Setting invalid exit code.')
            exit_code = 6
        if counter > 20:
            return None
        counter += 1
    if small:
        initial[:-1] /= la.norm(initial[:-1])
    else:
        initial /= la.norm(initial)
    if verbose:
        print('done! Exit code:', exit_code, '\n', initial)
        sys.stdout.flush()
    return initial

def __torus_dists (p, q):
    d = np.abs(p-q)
    d[d > PI] -= 2 * PI
    return la.norm(d, axis=1)

def get_sphere (points, orig_points, list_spheres, verbose, mode):
    N, d = points.shape
    N2 = len(max_linear_independent_subset(points))
    if N <= d or N2 < d:
        tmp = points.copy()
        sup_vec = tmp[-1]
        tmp = gram_schmidt(tmp[:-1] - sup_vec)
        n = len(tmp)
        if verbose:
            print('%d points with dimension %d. Making S^%d' % (N, d, n-1))
        tmp = np.vstack((tmp, 2 * arandom.rand(d - n, d) - 1))
        normals = gram_schmidt(tmp)[n:]
        sup_vec = np.einsum('ij,j->i', normals, sup_vec)
        if normals.shape[0] == 1 and sup_vec.shape[0] == 1:
            return Sphere1(normals[0], sup_vec[0])
        return Sphere(normals, sup_vec)
    N2 = len(max_linear_independent_subset(points))
    if mode == 'torus':
        projector_sphere = Sphere.unfold(list_spheres)
    """
    Function f: R^(d+1) -> R^N for which the Levenberg-Marquardt method is
    applied to identify the argument which minimizes the sum of squares.
    """
    def f (x):
        if abs(x[-1]) > 1:
            if verbose:
                print('Fail:', x[-1])
            return 2 * PI * np.ones(points.shape[0]) * abs(x[-1])
        return Sphere1(x[:-1] / la.norm(x[:-1]), x[-1]).distances(points)
    def f2 (x):
        return Sphere1(x / la.norm(x), 0).distances(points)
    def g (x):
        if abs(x[-1]) > 1:
            if verbose:
                print('Fail:', x[-1])
            return 2 * PI * np.ones(points.shape[0]) * abs(x[-1])
        feet = Sphere1(x[:-1] / la.norm(x[:-1]), x[-1]).foot_points(points)
        if projector_sphere:
            return __torus_dists(projector_sphere.unproject(feet), orig_points)
        return __torus_dists(feet, orig_points)
    tmp = 2 * arandom.rand(d + 1) - 1
    if mode == 'torus':
        tmp = fit(g, d, tmp, verbose)
    elif mode == 'great':
        tmp = fit(f2, d, tmp[:-1], verbose, False)
        if tmp is None:
            return None
        return Sphere1(tmp / la.norm(tmp), 0)
    else:
        tmp = fit(f, d, tmp, verbose)
    if tmp is None:
        return None
    if mode == 'scale':
        radii = np.arccos(np.abs(np.einsum('ij,j->i', points, tmp[:-1])))
        if compare_likelihoods(radii, d-1, verbose):
            tmp2 = tmp.copy()
            tmp = fit(f2, d, tmp[:-1], verbose, False)
            ang = np.arccos(np.einsum('i,i->', tmp, tmp2[:-1])) * DEG
            if verbose: print('CHANGED:', ang, abs(tmp2[-1]))
            return Sphere1(tmp / la.norm(tmp), 0)
    return Sphere1(tmp[:-1] / la.norm(tmp[:-1]), tmp[-1])

def pns_loop (points, degenerate=False, verbose=False, mode=None):
    list_spheres = []
    list_points = []
    list_dists = []
    orig_points = points.copy()
    while points.shape[1] > 2:
        # Do PNS
        list_spheres.append(get_sphere(points, orig_points, list_spheres, verbose, mode))
        if list_spheres[-1] is None:
            return None, None, None
        dists, feet = list_spheres[-1].distances(points, True)
        if degenerate:
            points_out = unfold_points(points, list_spheres[:-1])
            sphere_out = Sphere.unfold(list_spheres)
            # Determine alternative distances and projections
            dists, feet = alt_dist_loop(points_out, dists, feet, sphere_out,
                                        list_spheres[:-1], verbose)
        points = list_spheres[-1].project(feet)
        list_points.append(points)
        list_dists.append(dists * DEG)
    if points.shape[1] < 2:
        return None, None, None
    mean, residuals = nested_mean(points)
    list_spheres.append(None)
    list_points.append(mean)
    list_dists.append(residuals)
    return list_spheres, list_points, list_dists

################################################################################
######################   Calculation of circular means   #######################
################################################################################
"""
Calculation of circular means is based on the results of
Hotz, T., Huckemann, S.:
"Intrinsic Means on the Circle: Uniqueness, Locus and Asymptotics".
The Annals of the Institute of Mathematical Sciences, arXiv:1108.2141
To appear (2014)
"""

def center_of_mass (points_2d):
    phi, var = circular_mean(np.arctan2(points_2d[:,1], points_2d[:,0]))
    return np.array([math.cos(phi), math.sin(phi)]), sqrt(var) * DEG

def euclideanize (angles):
    return np.array([((5 * PI + a - circular_mean(a)[0]) % (2 * PI) - PI) * DEG
                     for a in angles])

def nested_mean (points_2d):
    phis = np.arctan2(points_2d[:,1], points_2d[:,0])
    mean, _ = circular_mean(phis)
    phis = (5 * PI + phis - mean) % (2 * PI) - PI
    return np.array([math.cos(mean), math.sin(mean)]), phis * DEG

def circular_mean (points):
    n = points.size
    mean0 = np.mean(points)
    var0 = np.var(points)
    sorted_points = np.sort(points)
    means = variances(mean0, var0, n, sorted_points)
    return means[np.argmin(means[:,1])]

def variances (mean0, var0, n, points):
    means = [(mean0 + 2 * PI * i / float(n)) for i in range(n)]
    means = [(x if x < PI else x - 2 * PI) for x in means]
    parts = [(sum(points) / n) if means[0] < 0 else 0]
    parts += [((sum(points[:i]) / i) if means[i] >= 0 else (sum(points[i:]) / (n-i)))
             for i in range(1, len(means))]
    #Formula (6) from Hotz, Huckemann
    means = [[means[i],
              var0 + ( 0 if i == 0 else
                       ((4 * PI * i / n) * (PI + parts[i] - mean0) -
                        (2 * PI * i / n)**2) if means[i] >= 0 else
                       ((4 * PI * (n - i) / n) * (PI - parts[i] + mean0) -
                        (2 * PI * (n - i) / n)**2) )]
             for i in range(len(means))]
    return np.array(means)

################################################################################
#############################   Likelihood Test   ##############################
################################################################################

def normalization (rho, sigma, d):
    def f (r): return (exp(-0.5*(r/sigma-rho)**2) + exp(-0.5*(r/sigma+rho)**2))
    try:
        return max(sys.float_info.min, quad(lambda r: sin(r)**(d-1) * f(r), 0, PI)[0])
    except:
        #print("WARNING: Integral did not converge for sigma ", sigma)
        return max(sys.float_info.min, sqrt(2 * PI) * sigma)

def compare_likelihoods (radii, d, verbose):
    mean = radii.mean()
    std = radii.std()
    def likelihood (x):
        penalty = 0.
        scale = 2. * x[0] * x[1] / PI
        if scale > 1:
            x[0] = 0.5 * PI / x[1]
            penalty = scale
        out = np.sum(log(normalization(x[0],x[1],d)) + 0.5*(radii/x[1]-x[0])**2 -
                     np.log(1 + np.exp(- 2. * x[0] * radii / x[1]))) + penalty
        return out
    def likelihood_alt (x):
        penalty = 0.
        scale = 2. * x[0] * x[1] / PI
        if scale > 1:
            x[0] = 0.5 * PI / x[1]
            penalty = scale
        out = np.sum(log(x[1]) + 0.5*(radii/x[1]-x[0])**2 -
                     np.log(1 + np.exp(- 2. * x[0] * radii / x[1]))) + penalty
        return out
    def likelihood_null (x):
        return likelihood(np.array([1,x]))
    mle = opt.minimize(likelihood, np.array([max(mean/std, 1.), std]),
                       method=('L-BFGS-B' if OLD_SCIPY else None),
                       bounds=((0,PI*1e3),(max(1e-3,0.25*std),0.5*PI))).x
    alt_mle = opt.minimize(likelihood_alt, np.array([max(mean/std, 1.), std]),
                           method=('L-BFGS-B' if OLD_SCIPY else None),
                           bounds=((0,PI*1e3),(max(1e-3,0.25*std),0.5*PI))).x
    if verbose:
        print('Mean: %.3f %.3f' % (mean, std),
              'MLE: %.3f %.3f' % (mle[0]*mle[1], mle[1]),
              'JMD MLE: %.3f %.3f' % (alt_mle[0], alt_mle[1]))
    if mle[0] < 1:
        return True
    mle_null = opt.minimize(likelihood_null, 1,
                            method=('L-BFGS-B' if OLD_SCIPY else None),
                            bounds=((mle[1],0.5*PI),)).x
    chi2 = 1 - stat.chi2.cdf(2 * (likelihood_null(mle_null[0]) - likelihood(mle)), 1)
    if (verbose and (chi2 > 0)): print('Chi-square: ', chi2)
    return chi2 > 0.05

def __likelihoods (radii, d):
    mean = radii.mean()
    std = radii.std()
    def likelihood (x):
        penalty = 0.
        scale = 2. * x[0] * x[1] / PI
        if scale > 1:
            x[0] = 0.5 * PI / x[1]
            penalty = scale
        out = np.sum(log(normalization(x[0],x[1],d)) + 0.5*(radii/x[1]-x[0])**2 -
                     np.log(1 + np.exp(- 2. * x[0] * radii / x[1]))) + penalty
        return out
    def likelihood_alt (x):
        penalty = 0.
        scale = 2. * x[0] * x[1] / PI
        if scale > 1:
            x[0] = 0.5 * PI / x[1]
            penalty = scale
        out = np.sum(log(x[1]) + 0.5*(radii/x[1]-x[0])**2 -
                     np.log(1 + np.exp(- 2. * x[0] * radii / x[1]))) + penalty
        return out
    def likelihood_null (x):
        return likelihood(np.array([1,x]))
    mle = opt.minimize(likelihood, np.array([max(mean/std, 1.), std]),
                       method=('L-BFGS-B' if OLD_SCIPY else None),
                       bounds=((0,PI*1e3),(max(1e-3,0.25*std),0.5*PI))).x
    alt_mle = opt.minimize(likelihood_alt, np.array([max(mean/std, 1.), std]),
                           method=('L-BFGS-B' if OLD_SCIPY else None),
                           bounds=((0,PI*1e3),(max(1e-3,0.25*std),0.5*PI))).x
    if mle[0] < 1:
        return mle[0], 1, alt_mle[0]
    mle_null = opt.minimize(likelihood_null, 1,
                            method=('L-BFGS-B' if OLD_SCIPY else None),
                            bounds=((mle[1],0.5*PI),)).x
    chi2 = 1 - stat.chi2.cdf(2 * (likelihood_null(mle_null[0]) - likelihood(mle)), 1)
    return mle[0], chi2, alt_mle[0]

def test_tests (points):
    N, d = points.shape
    """
    Function f: R^(d+1) -> R^N for which the Levenberg-Marquardt method is
    applied to identify the argument which minimizes the sum of squares.
    """
    def f (x):
        if abs(x[-1]) > 1:
            return 2 * PI * np.ones(points.shape[0]) * abs(x[-1])
        return Sphere1(x[:-1] / la.norm(x[:-1]), x[-1]).distances(points)
    tmp = 2 * arandom.rand(d + 1) - 1
    tmp = fit(f, d, tmp, False)
    if tmp is None:
        return None
    radii = np.arccos(np.abs(np.einsum('ij,j->i', points, tmp[:-1])))
    return __likelihoods(radii, d-1)

################################################################################
#############################   Mean of Spheres   ##############################
################################################################################

def two_sample_test (data1, data2, samples=100, verbose=False, mode='scale'):
    n1, d = data1.shape
    n2, d2 = data2.shape
    if d != d2:
        print('Data have different dimension!')
        return None
    if verbose: print('Bootstrapping for data set 1.')
    samples1 = bootstrap(data1, samples, verbose, mode)
    if verbose: print('Bootstrapping for data set 2.')
    samples2 = bootstrap(data2, samples, verbose, mode)
    probs = [yao_hotelling(np.array(samples1[0]), np.array(samples2[0]), n1, n2)]
    for i in range(1,len(samples1)):
        probs.append(__test(samples1[i], samples2[i], n1, n2, verbose))
    return probs


def bootstrap (data, s, verbose, mode):
    n, d = data.shape
    out = [[] for _ in range(d - 1)]
    while len(out[0]) < s:
        bdata = data[arandom.choice(np.array(range(n)), 2 * n * s)].reshape((2 * s, n, d))
        for x in bdata:
            spheres, projected_points, distances = pns_loop(x, mode=mode,
                                                            verbose=verbose)
            if spheres is None or len(spheres) < len(out):
                if verbose: print("Low dimensional sample. Skipping...")
                continue
            out[0].append(unfold_points(projected_points[-1][np.newaxis,:],
                                        spheres[:-1])[0])
            for i in range(1,len(spheres)):
                out[i].append(Sphere.unfold(spheres[:-i]))
            if len(out[0]) == s:
                return out
    return None

def __test (sample1, sample2, n1, n2, verbose):
    if verbose: print('Determine chart...')
    mean, vecs = get_mean_sphere(sample1 + sample2, verbose)
    prob = yao_hotelling(vecs[:len(sample1)], vecs[len(sample1):], n1, n2)
    return prob

def yao_hotelling (x1, x2, n1, n2):
    n,d = x1.shape
    cov1 = np.cov(x1.T) * n1 / (n1-1); cov2 = np.cov(x2.T) * n2 / (n2-1)
    inv_cov = la.inv(cov1 + cov2)
    mu_diff = np.mean(x1, axis=0) - np.mean(x2, axis=0)
    t2 = np.einsum('i,ij,j->', mu_diff, inv_cov, mu_diff)
    tmp = np.einsum('i,ij->j', mu_diff, inv_cov)
    m1 = np.einsum('i,ij,j->', tmp, cov1, tmp)
    m2 = np.einsum('i,ij,j->', tmp, cov2, tmp)
    n = t2**2 / (m1**2 / (n1 - 1) + m2**2 / (n2 - 1))
    return 1 - stat.f.cdf((n+1-d) * t2 / (n*d), d, n + 1 - d)

def get_mean_sphere (sphere_list, verbose):
    if len(sphere_list) < 1:
        return None
    if len(sphere_list) == 1:
        return sphere_list[0]
    descriptors = np.array([s.descriptor() for s in sphere_list])
    shape = descriptors[0].shape
    dimension = ((shape[0] - 1) if (len(shape) == 1) else
                 (shape[0] * (shape[1] - shape[0])))
    tmp = None
    old_dist = 2e3
    new_dist = 1e3
    mean = None
    while (old_dist - new_dist > 1e-5):
        old_dist = new_dist
        if tmp is None:
            mean = descriptors[arandom.randint(0, len(descriptors), 1)[0]]
        else:
            mean = np.mean(np.array([a[1] for a in tmp]), axis = 0)
        if len(mean.shape) == 1:
            mean[:-1] /= la.norm(mean[:-1])
        else:
            mean[:,:-1] = gram_schmidt(mean[:,:-1])
        tmp = [__descriptor_distance(mean, a) for a in descriptors]
        new_dist = la.norm(np.array([a[0] for a in tmp]))
        if verbose: print(mean, new_dist)
        sys.stdout.flush()
    residuals = np.array([(a[1]-mean).flatten() for a in tmp])
    val, vec = la.eig(np.cov(residuals.T))
    order = (val.argsort()[::-1])[:dimension]
    val, vec = np.real(val[order]), np.real((vec.T)[order])
    if verbose: print(val)
    return mean, np.einsum('ik,jk->ij', residuals, vec)

def __descriptor_distance (a, b):
    #print('Calculating distance!')
    sys.stdout.flush()
    if len(a.shape) == 1:
        db, b2 = la.norm(a-b), b
        dc, c2 = la.norm(a+b), -b
    else:
        c = b.copy()
        c[[0,1]] = c[[1,0]]
        db, b2 = __min_dist(a,b)
        dc, c2 = __min_dist(a,c)
    return (db, b2) if db < dc else (dc, c2)

def __min_dist (a, b):
    z = np.einsum('ij,kj->ik', a, b)
    u,s,v = la.svd(z)
    r = np.einsum('ji,jk->ik', u, v)
    best_b = np.real(np.einsum('ij,jk->ik',r,b))
    return la.norm((a - best_b).flatten()), best_b

def __fit (f, d):
    warnings.filterwarnings('error')
    initial = arandom.rand(d)
    exit_code = 6
    fails = -1
    counter = 0
    while exit_code > 1 and la.norm(f(initial)) > EPS:
        fails += 1
        if fails > 3 or exit_code == 6:
            initial = arandom.rand(d)
            fails = 0
        try:
            initial, exit_code = opt.leastsq(f, initial)
        except:
            exit_code = 6
            counter += 1
        if counter > 20:
            return None
    return initial

################################################################################
###########################   Auxiliary functions   ############################
################################################################################

def alt_dist_loop (points, dists, feet, sphere, list_spheres, verbose):
    """
    Determine alternative distances and projections to the PNS sphere for
    points that are sufficiently close to the degenerate sphere.
    """
    dists_deg = deg_distances(points, 0)
    n_close = 0
    n_changed = 0
    for i in range(len(dists)):
        if dists[i] > dists_deg[i]:
            n_close += 1
            tmp = alt_distance(points[i], dists[i], sphere)
            if tmp[0] + EPS < dists[i]:
                n_changed += 1
                if list_spheres:
                    dists[i], foot = tmp
                    feet[i] = fold_points(foot, list_spheres)
                else:
                    dists[i], feet[i] = tmp
    if verbose: print(n_changed, 'of', n_close, 'distances changed.')
    sys.stdout.flush()
    return dists, feet

def alt_distance (p_data, dist, pns_sphere):
    """
    This function is called only for points close to the degenerate sphere.
    p_data     : data point
    dist       : distance of p_data to the PNS sphere
    pns_sphere : the PNS sphere
    """
    n_max = len(p_data) - 2
    min_dist = dist
    best_foot = None
    for i in range(n_max):
        dist_deg, q_data = deg_distance(p_data, i, True)
        if dist_deg > min_dist:
            return min_dist, best_foot
        # Find best passage through degenerate sphere.
        def d (q, with_point = False):
            q /= la.norm(q)
            q = np.hstack((q,np.array([0]*(2+i))))
            q_bar = q.copy()
            q_bar[-3-i] *= -1
            p_pns = pns_sphere.foot_points(as_matrix(q_bar))
            if with_point:
                return bow(p_data, q) + bow(q_bar, as_vector(p_pns)), p_pns
            return bow(p_data, q) + bow(q_bar, as_vector(p_pns))
        alt_dist, p_pns = d(opt.minimize(d, q_data[:-2-i]).x, True)
        if alt_dist < min_dist:
            min_dist = alt_dist
            best_foot = p_pns
    return min_dist, best_foot

def fold_points (points, list_spheres):
    out = points.copy()
    for sphere in list_spheres:
        out = sphere.project(out)
    return out

def unfold_points (points, list_spheres):
    out = points.copy()
    for sphere in reversed(list_spheres):
        out = sphere.unproject(out)
    return out

def bow (p, q):
    prod = np.einsum('i,i->', p, q)
    if prod > 1: prod = 1
    if prod < -1: prod = -1
    return acos(prod)

def deg_distances (points, codim, with_feet = False):
    if with_feet:
        tmp = points.copy()
        tmp[:,-2-codim:] = 0
        feet = np.einsum('ij,i->ij', tmp, 1/la.norm(tmp, axis=1))
        dists = np.arccos(np.einsum('ij,ij->i', points, feet).clip(-1, 1))
        return dists, feet
    return np.arccos(la.norm(points[:,:-2-codim], axis=1).clip(-1, 1))

def deg_distance (point, codim, with_foot = False):
    if with_foot:
        foot = point.copy()
        foot[-2-codim:] = 0
        foot /= la.norm(foot)
        dist = acos(max(min(np.einsum('i,i->', point, foot), 1), -1))
        return dist, foot
    return np.acos(la.norm(point[:-2-codim]))

def as_matrix (vector):
    return vector.reshape(1, len(vector))

def as_vector (matrix):
    return matrix.reshape(matrix.shape[1])

def gram_schmidt (vectors):
    if vectors.shape[0] == 1:
        return vectors
    tmp = max_linear_independent_subset(vectors)
    tmp = np.einsum('ij,i->ij', tmp, 1/la.norm(tmp, axis=1))
    if len(tmp) == 0:
        return None
    out = tmp[[0]]
    for i in range(1, len(tmp)):
        new = (tmp[i] - np.einsum('ij,ik,j->k', out, out, tmp[i]))
        out = np.vstack((out, new/la.norm(new)))
    return out

def orthogonal (vector):
    return gram_schmidt(np.vstack((vector, arandom.rand(len(vector)))))[-1]

def max_linear_independent_subset (vectors):
    counter = 0
    out = None
    for i in range(len(vectors)):
        if la.norm(vectors[i]) < EPS:
            pass
        elif counter == 0:
            out = vectors[i][np.newaxis,:]
            counter = 1
        else:
            tmp = np.vstack((out, vectors[i]))
            if la.matrix_rank(tmp, EPS) > counter:
                out = tmp
                counter += 1
    return out

def array2csv (array, form = '%.6f', s=''):
    if len(array.shape) == 1:
        return s + ','.join(form % item for item in array) + '\n'
    if len(array.shape) == 2:
        return s + ('\n'+s).join(','.join(form % i for i in k) for k in array) + '\n'
    return None

################################################################################
#########################   Auxiliary sphere classes   #########################
################################################################################

"""
This class describes a sub-sphere with codimension 1 inside a sphere.
normal : center of the sub-sphere on the sphere
center : center of the sub-sphere in embedding space
"""
class Sphere1 (object):
    def __init__ (self, normal, height):
        self.normal = normal
        self.height = max(-1, min(1, height))
        if self.height < 0:
            self.height *= -1
            self.normal *= -1
        self.center = self.height * self.normal
        self.radius = acos(self.height)
        self.flat_radius = sqrt(1 - self.height**2)

    def on_plane (self, point):
        return np.abs(np.inner((point - self.center), self.normal)) < EPS

    def foot_points (self, points):
        tmp = points - np.einsum('j,k,lj->lk', self.normal, self.normal, points)
        if np.any(la.norm(tmp, axis=1) == 0):
            return None
        return self.center + self.flat_radius * np.einsum('ij,i->ij', tmp,
                                                           1/la.norm(tmp, axis=1))

    def distances (self, points, with_feet = False):
        dists = (self.radius - np.arccos(np.einsum('ij,j->i', points,
                                                   self.normal).clip(-1, 1)))
        if with_feet:
            feet = self.foot_points(points)
            return dists, feet
        return dists

    def calculate_projection_matrix (self):
        self.projection_matrix = np.eye(len(self.normal))
        self.projection_matrix[0] = self.normal
        self.projection_matrix = gram_schmidt(self.projection_matrix)[1:]

    def project (self, points):
        if not hasattr(self, 'projection_matrix'):
            self.calculate_projection_matrix()
        return np.einsum('ik,jk->ij', self.foot_points(points),
                          self.projection_matrix) / self.flat_radius

    def unproject (self, points):
        if not hasattr(self, 'projection_matrix'):
            print('WARNING: This sphere has not been used for projection!')
            self.calculate_projection_matrix()
        return (np.einsum('ik,kj->ij', points, self.projection_matrix) *
                self.flat_radius + self.center)

    def descriptor (self):
        return np.hstack((self.normal, np.array([self.height])))

    def __str__ (self):
        if not hasattr(self, 'projection_matrix'):
            self.calculate_projection_matrix()
        return Sphere.string(self.flat_radius, self.center, self.projection_matrix)


"""
This class describes a sub-sphere with arbitrary codimension c inside a sphere.
As a sub-sphere of codimension c is uniquely defined by a hyperplane of
codimension c, the input data must specify such a hyperplane.
normals  : An array of c orthonormal vectors, that are orthogonal on the
           hyperplane.
position : A c-dimensional vector specifying the position of the hyperplane in
           terms of the normals.
"""
class Sphere (object):
    def __init__ (self, normals, position):
        if normals.shape[0] != len(position):
            print('INVALID HYPERPLANE!\n', normals, '\n', position)
            return
        if (la.norm(normals - gram_schmidt(normals)) > EPS):
            print('NORMALS ARE NOT ORTHONORMAL!',
                  la.norm(normals - gram_schmidt(normals)), '\n', normals)
            return
        self.normals = normals
        height = la.norm(position)
        if height > 1:
            print('INVALID POSITION!\n', position)
            return
        self.center = np.einsum('ij,i->j', self.normals, position)
        self.position = position
        self.flat_radius = sqrt(1 - height**2)

    @classmethod
    def unfold (cls, list_of_spheres):
        """
        Transform a list of n codimension 1 spheres into a codimension n sphere.
        """
        if list_of_spheres is None or len(list_of_spheres) == 0:
            return None
        if len(list_of_spheres) == 1:
            return list_of_spheres[0]
        normals = []
        position = []
        for i, sphere in enumerate(list_of_spheres):
            try:
                normal = sphere.normal.copy()
                height = sphere.height
                for sphere in reversed(list_of_spheres[:i]):
                    normal = np.einsum('k,kj->j', normal, sphere.projection_matrix)
                    height *= sphere.flat_radius
                normals.append(normal)
                position.append(height)
            except:
                if i > 0:
                    return None
                print("Problem unfolding. Trying Sphere instead of Sphere1.")
                normals = [n for n in sphere.normals.copy()]
                position = [p for p in sphere.position.copy()]
        return Sphere(np.array(normals), np.array(position))

    @classmethod
    def string(cls, r, c, m):
        return '%%radius\n%%%.6f\n%%center\n%s%%matrix\n%s' % (r, array2csv(c),
                                                               array2csv(m))

    def foot_points (self, points):
        tmp = points - np.einsum('ij,ik,lj->lk', self.normals, self.normals, points)
        if np.any(la.norm(tmp, axis=1) == 0):
            return None
        return self.center + self.flat_radius * np.einsum('ij,i->ij', tmp,
                                                           1/la.norm(tmp, axis=1))

    def distances (self, points, with_feet = False):
        feet = self.foot_points(points)
        dists = np.arccos(np.einsum('ij,ij->i', points, feet).clip(-1, 1))
        if with_feet:
            return dists, feet
        return dists

    def calculate_projection_matrix (self):
        self.projection_matrix = np.eye(self.normals.shape[1])
        self.projection_matrix[:len(self.normals)] = self.normals
        self.projection_matrix = gram_schmidt(self.projection_matrix)[len(self.normals):]

    def project (self, points):
        if not hasattr(self, 'projection_matrix'):
            self.calculate_projection_matrix()
        return np.einsum('ik,jk->ij', self.foot_points(points),
                          self.projection_matrix) / self.flat_radius

    def unproject (self, points):
        if not hasattr(self, 'projection_matrix'):
            print('WARNING: This sphere has not been used for projection!')
            self.calculate_projection_matrix()
        return (np.einsum('ik,kj->ij', points, self.projection_matrix) *
                self.flat_radius + self.center)

    def descriptor (self):
        return np.hstack((self.normals, np.einsum('ij,j->i', self.normals,
                                                  self.center)[:,np.newaxis]))

    def __str__ (self):
        if not hasattr(self, 'projection_matrix'):
            self.calculate_projection_matrix()
        return Sphere.string(self.flat_radius, self.center, self.projection_matrix)
