package clustering;
/*
 * This file is part of CANE - A Clustering Algorithm for Non-Euclidean spaces.
 *
 * Copyright (C) 2014 Benjamin Eltzner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Random;

import org.apache.commons.math3.analysis.MultivariateMatrixFunction;
import org.apache.commons.math3.analysis.MultivariateVectorFunction;
import org.apache.commons.math3.analysis.UnivariateVectorFunction;
import org.apache.commons.math3.analysis.differentiation.DerivativeStructure;
import org.apache.commons.math3.analysis.differentiation.FiniteDifferencesDifferentiator;
import org.apache.commons.math3.analysis.differentiation.UnivariateDifferentiableVectorFunction;
import org.apache.commons.math3.exception.TooManyIterationsException;
import org.apache.commons.math3.fitting.leastsquares.LeastSquaresFactory;
import org.apache.commons.math3.fitting.leastsquares.LeastSquaresOptimizer;
import org.apache.commons.math3.fitting.leastsquares.LeastSquaresProblem;
import org.apache.commons.math3.fitting.leastsquares.LevenbergMarquardtOptimizer;
import org.apache.commons.math3.fitting.leastsquares.MultivariateJacobianFunction;
import org.apache.commons.math3.linear.ArrayRealVector;
import org.apache.commons.math3.linear.RealVector;

/**
 * This class wraps a SimpleClusterable and implements the Clusterable
 * interface. The Levenberg-Marquardt algorithm is used to determine
 * centroid positions via moveToCenter() and a random splitting procedure
 * is applied for the children() method.
 */
public class ClusterableFromSimple implements Clusterable
{

    private SimpleClusterable m_simple;

    public ClusterableFromSimple(SimpleClusterable simple)
    {
        m_simple = simple.clone();
    }

    @Override
    public String exportCenter()
    {
        return m_simple.exportCenter();
    }

    @Override
    public int label() throws NotImplementedException
    {
        return m_simple.label();
    }

    @Override
    public int dimension()
    {
        return m_simple.dimension();
    }

    @Override
    public long distanceSquare(Clusterable that)
    {
        if (that == null || !that.getClass().equals(this.getClass()))
        {
            return Long.MAX_VALUE;
        }
        ClusterableFromSimple other = (ClusterableFromSimple) that;
        if (!m_simple.getClass().equals(other.m_simple.getClass()))
        {
            return Long.MAX_VALUE;
        }

        double d = m_simple.distance(other.m_simple);
        return Math.round(d * d);
    }

    @Override
    public Clusterable moveToCenter(List<Clusterable> cluster)
    {
        // Quick escape for single data points.
        if (cluster != null && cluster.size() == 1 &&
            cluster.get(0).getClass().equals(ClusterableFromSimple.class))
        {
            m_simple.set(((ClusterableFromSimple)cluster.get(0)).m_simple.toArray(), true);
        }
        m_simple.set(getCenter(cluster), true);
        return this;
    }

    private double[] getCenter(List<Clusterable> cluster)
    {
        if (cluster == null)
        {
            System.err.println("Center has no data!");
            return m_simple.toArray();
        }
        Distances f = new Distances(cluster);
        MultivariateMatrixFunction df = new FiniteDifferencesJacobian(f);
        MultivariateJacobianFunction model = LeastSquaresFactory.model(f, df);
        RealVector observed = new ArrayRealVector(new double[f.m_output_size]);
        RealVector start = new ArrayRealVector(m_simple.toArray());
        int steps = 50;
        LevenbergMarquardtOptimizer opt = new LevenbergMarquardtOptimizer();
        LeastSquaresOptimizer.Optimum output = null;
        while (output == null)
        {
            LeastSquaresProblem problem = LeastSquaresFactory.create(model, observed,
                                                                     start, null,
                                                                     Integer.MAX_VALUE,
                                                                     steps);
            try
            {
                output = opt.optimize(problem);
            }
            catch(TooManyIterationsException e)
            {
                System.err.println("Levenberg-Marquardt needed more than " + steps + " steps!");
                double wiggle = start.getNorm() / 100.;
                for (int i = 0; i < start.getDimension(); i++)
                {
                    start.addToEntry(i, wiggle * (2 * new Random().nextDouble() - 1));
                }
                start.unitize();
                start.mapMultiply(100 * wiggle);
                steps += 50;
            }
        }
        
        return output.getPoint().toArray();
    }

    @Override
    public List<Clusterable> children(List<Clusterable> cluster)
    {
        Random random = new Random();
        int d = m_simple.dimension();
        double[] shift = new double[d];
        double shift_norm = 0;
        for (int i = 0; i < d; i++)
        {
            shift[i] = random.nextDouble();
            shift_norm += shift[i] * shift[i];
        }
        Clusterable tmp = cluster.get(random.nextInt(cluster.size()));
        if (tmp == null || !tmp.getClass().equals(this.getClass()))
        {
            System.err.println("Incompatible types: Clusterable is not ClusterableFromSimple.");
            return null;
        }
        ClusterableFromSimple random_point = (ClusterableFromSimple) tmp;
        if (!m_simple.getClass().equals(random_point.m_simple.getClass()))
        {
            System.err.println("Incompatible types: SimpleClusterable types incompatible.");
            return null;
        }

        shift_norm = Math.sqrt(shift_norm) / m_simple.distance(random_point.m_simple);
        double[] shift_minus = new double[d];
        for (int i = 0; i < d; i++)
        {
            shift[i] /= shift_norm;
            shift_minus[i] = - shift[i];
        }

        ClusterableFromSimple child1 = (ClusterableFromSimple)this.clone();
        ClusterableFromSimple child2 = (ClusterableFromSimple)this.clone();
        List<Clusterable> out = new ArrayList<>();
        out.add(child1.move(shift));
        out.add(child2.move(shift_minus));
        return out;
    }

    private Clusterable move(double[] shift)
    {
        double[] value = m_simple.toArray();
        for (int i = 0; i < value.length; i++)
        {
            value[i] += shift[i];
        }
        m_simple.set(value, false);
        return this;
    }

    @Override
    public double[] projection(List<Clusterable> direction, List<Clusterable> cluster)
           throws NotImplementedException
    {
        return m_simple.projection(direction, cluster);
    }

    @Override
    public String toString()
    {
        return m_simple.toString();
    }

    @Override
    public Clusterable clone()
    {
        return new ClusterableFromSimple(m_simple.clone());
    }

    @Override
    public int hashCode()
    {
        return m_simple.hashCode();
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        if (obj == null || getClass() != obj.getClass())
        {
            return false;
        }
        ClusterableFromSimple other = (ClusterableFromSimple) obj;
        return m_simple.equals(other.m_simple);
    }

    private class Distances implements FunctionVectorToVector
    {
        private List<SimpleClusterable> m_cluster;
        private int m_input_size,
                    m_output_size;

        private Distances(List<? extends Clusterable> cluster)
        {
            List<SimpleClusterable> list = new ArrayList<>();
            for (Clusterable that : cluster)
            {
                if (list != null && that != null &&
                    that.getClass().equals(ClusterableFromSimple.class))
                {
                    ClusterableFromSimple other = (ClusterableFromSimple) that;
                    if (m_simple.getClass().equals(other.m_simple.getClass()))
                    {
                        list.add(other.m_simple);
                        if (m_input_size == 0)
                        {
                            m_input_size = other.m_simple.dimension();
                        }
                    }
                    else
                    {
                        list = null;
                    }
                }
                else
                {
                    list = null;
                }
            }
            m_cluster = list;
            m_input_size  = (m_cluster == null ? 0 : m_input_size);
            m_output_size = (m_cluster == null ? 0 : m_cluster.size());
        }

        @Override
        public double[] value(double[] center) throws IllegalArgumentException
        {
            if (m_output_size == 0 || m_cluster == null || m_cluster.isEmpty())
            {
                return null;
            }
            double[] value = new double[m_output_size];
            //double sum_sq_dist = 0.0;
            for (int i = 0; i < m_output_size; i++)
            {
                value[i] = m_cluster.get(i).distance(center);
                //sum_sq_dist += value[i] * value[i];
            }
            //System.out.println("Center: " + Arrays.toString(center) + ", sum of squares: " + sum_sq_dist);
            return value;
        }

        @Override
        public int inputDimension()
        {
            return m_input_size;
        }

        @Override
        public int outputDimension()
        {
            return m_output_size;
        }
    }

    private class FiniteDifferencesJacobian implements MultivariateMatrixFunction
    {
        private MultivariateVectorFunction m_function;
        private int m_input_dimension,
                    m_output_dimension;
        
        private FiniteDifferencesJacobian(FunctionVectorToVector f)
        {
            m_function = f;
            m_input_dimension = f.inputDimension();
            m_output_dimension = f.outputDimension();
        }

        @Override
        public double[][] value(double[] x) throws IllegalArgumentException
        {
            double[][] output = new double[m_output_dimension][m_input_dimension];
            Restriction f_res = new Restriction(m_function, x);
            for (int i = 0; i < x.length; i++)
            {
                f_res.setPosition(i);
                FiniteDifferencesDifferentiator differ = new FiniteDifferencesDifferentiator(2, 1);
                double[] df_i = (new Derivative(differ.differentiate(f_res))).value(x[i]);
                for (int j = 0; j < m_output_dimension; j++)
                {
                    output[j][i] = df_i[j];
                }
            }
            //System.out.println("Jacobian at point: " + Arrays.toString(x));
            return output;
        }

        private class Restriction implements UnivariateVectorFunction
        {
            private double[] m_point;
            private int m_position;
            private MultivariateVectorFunction m_f;

            private Restriction(MultivariateVectorFunction f, double[] point)
            {
                m_f = f;
                m_point = point;
            }

            private void setPosition (int position)
            {
                m_position = position;
            }
            
            @Override
            public double[] value(double arg0)
            {
                double[] x = Arrays.copyOf(m_point, m_point.length);
                x[m_position] = arg0;
                return m_f.value(x);
            }
        }

        private class Derivative implements UnivariateVectorFunction
        {
            private UnivariateDifferentiableVectorFunction m_f;

            private Derivative(UnivariateDifferentiableVectorFunction f)
            {
                m_f = f;
            }

            @Override
            public double[] value(double arg0)
            {
                DerivativeStructure x = new DerivativeStructure(1, 1, 0, arg0);
                DerivativeStructure[] tmp = m_f.value(x);
                double[] output = new double[tmp.length];
                for (int i = 0; i < tmp.length; i++)
                {
                    output[i] = tmp[i].getPartialDerivative(1);
                }
                return output;
            }
        }
    }

}
