package clustering;
/*
 * This file is part of CANE - A Clustering Algorithm for Non-Euclidean spaces.
 *
 * Copyright (C) 2014 Benjamin Eltzner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import org.apache.commons.math3.special.Erf;

/**
 * This class provides facilities for clustering of any Clusterable objects.
 */
public class Clusterer
{

    private List<Segmentation> m_segmentations;
    protected List<Clusterable> m_data_points;
    public static final int BIC = 0, GAUSS = 1, SILHOUETTE = 2, MIN_DIST = 3;
    private static final double MF = 1e6;
    
    /**
     * Initiate the class by passing it a list of data points.
     */
    public Clusterer(List<Clusterable> list)
    {
        m_data_points = list;
        m_segmentations = new ArrayList<>();
    }

    /**
     * This is the major control function of the class. It controls the clustering.
     * @param centers    A list of initial centroids.
     * @param max        The maximum number of centroids at which the algorithm is stopped.
     * @param type       The method used to determine the optimal number of clusters.
     * @param parameters Mode specific parameters.
     * @return           A map of centroids and their corresponding clusters.
     */
    public Map<Clusterable, List<Clusterable>> cluster(List<Clusterable> centers,
                                                       int max, final int type,
                                                       Number... parameters)
           throws NotImplementedException
    {
        if (centers == null || centers.isEmpty())
        {
            System.err.println("FilTracker Clusterer.cluster: " +
            		           "No valid initial center given.");
            return null;
        }

        int min = 0;
        long min_dist = 0,
             max_silhouette = Long.MIN_VALUE;
        long[][] distance_matrix = null;
        if (parameters != null &&  parameters.length > 0)
        {
            switch (type)
            {
            case SILHOUETTE:
                if (parameters.length > 1 && parameters[1].intValue() > 0)
                {
                    int n = m_data_points.size();
                    distance_matrix = new long[n][];
                    for (int i = 0; i < n; i++)
                    {
                        Clusterable tmp = m_data_points.get(i);
                        distance_matrix[i] = new long[i+1];
                        for (int j = 0; j <= i; j++)
                        {
                            distance_matrix[i][j] = Math.round(Math.sqrt(
                                                    tmp.distanceSquare(
                                                    m_data_points.get(j))));
                        }
                    }
                }
            case BIC:
                min = parameters[0].intValue();
                min = (min > max ? max : min);
                break;
            case MIN_DIST:
                min_dist = parameters[0].intValue();
                break;
            default:
                break;
            }
        }

        
        boolean live = true;

        while (live)
        {
            while (centers.size() < min)
            {
                Clusterable to_split = centers.remove((new Random()).nextInt(centers.size()));
                centers.addAll(to_split.children(m_data_points));
            }
            // K-Means over all centers.
            Segmentation cluster_segmentation = kMeans(centers, m_data_points, 0);
            if (cluster_segmentation == null || cluster_segmentation.isEmpty())
            {
                System.err.println("FilTracker Clusterer.cluster: Global clustering failed.");
                return null;
            }
            // This should work, because cluster_segmentation is defined anew on each run.
            m_segmentations.add(cluster_segmentation);

            if (type != SILHOUETTE)
            {
             // Check whether adding centers improves approximation quality.
                centers = new ArrayList<>();
                for (Cluster cluster : cluster_segmentation.m_clusters)
                {
                    Segmentation parent = new Segmentation(cluster);
                    Segmentation children = kMeans(cluster.m_center.children(cluster.m_data),
                                                   cluster.m_data, 0);
                    if (type == MIN_DIST && children != null && children.size() == 2)
                    {
                        List<Clusterable> tmp = children.centers();
                        if (tmp.get(0).distanceSquare(tmp.get(1)) >= min_dist * min_dist)
                        {
                            parent = null;
                        }
                    }

                    // Kill parent or children.
                    centers.addAll(split_test(parent, children, type) ? children.centers() : parent.centers());
                }

                // Break if adding centers does not enhance quality.
                if (centers.size() == cluster_segmentation.size() || centers.size() > max)
                {
                    if (centers.size() < min)
                    {
                        Clusterable to_split = centers.remove((new Random()).nextInt(centers.size()));
                        centers.addAll(to_split.children(cluster_segmentation.dataForCenter(to_split)));
                    }
                    else
                    {
                        live = false;
                    }
                }
                else if (centers.size() < cluster_segmentation.size())
                {
                    System.err.println("FilTracker Clusterer.cluster: " +
                                       "Refinement step lost cluster centers. This should not happen.");
                    return null;
                }
            }
            else if (type == SILHOUETTE)
            {
                // Break if maximum number of centers is reached or
                // new silhouette is too small.
                if (cluster_segmentation.size() >= Math.max(2, min))
                {
                    long new_silhouette = calculateSilhouette(cluster_segmentation,
                                                              distance_matrix);
                    System.out.println("silhouette is " + new_silhouette);
                    cluster_segmentation.m_score = new_silhouette;
                    if (new_silhouette < max_silhouette * 0.2)
                    {
                        live = false;
                    }
                    if (new_silhouette > max_silhouette)
                    {
                        max_silhouette = new_silhouette;
                    }
                }
                
                if (cluster_segmentation.size() >= max)
                {
                    live = false;
                }

                centers = cluster_segmentation.centers();
                Clusterable to_split = centers.remove((new Random()).nextInt(centers.size()));
                centers.addAll(to_split.children(cluster_segmentation.dataForCenter(to_split)));

                if (centers.size() < cluster_segmentation.size())
                {
                    System.err.println("FilTracker Clusterer.cluster: " +
                                       "Refinement step lost cluster centers. This should not happen.");
                    return null;
                }
            }
        }

        if (type == BIC)
        {
            // Find overall best fit and fill gaps around it.
            live = true;
            Segmentation best = null;
            while (live)
            {
                long max_score = Long.MIN_VALUE;
                best = null;
                for (Segmentation segmentation : m_segmentations)
                {
                    long score = scoreBIC(segmentation);
                    System.out.println(segmentation.size() + " clusters score " + score);
                    if (score > max_score)
                    {
                        max_score = score;
                        best = segmentation;
                    }
                }

                boolean lower_neighbor = false;
                for (Segmentation seg : m_segmentations)
                {
                    lower_neighbor = lower_neighbor || (seg.size() == best.size() - 1);
                }
                if (!lower_neighbor)
                {
                    System.out.println("Clustering with " + best.size() + " - 1 centroids missing.");
                    List<Clusterable> less_centers = best.centers();
                    less_centers.remove((new Random()).nextInt(less_centers.size()));
                    Segmentation segmentation = kMeans(less_centers, m_data_points, 0);
                    if (segmentation == null || segmentation.isEmpty())
                    {
                        System.err.println("FilTracker Clusterer.cluster: Global clustering failed.");
                        return null;
                    }
                    // This should work, because cluster_segmentation is defined anew on each run.
                    m_segmentations.add(segmentation);
                }

                boolean upper_neighbor = false;
                for (Segmentation seg : m_segmentations)
                {
                    upper_neighbor = upper_neighbor || (seg.size() == best.size() + 1);
                }
                if (!upper_neighbor)
                {
                    System.out.println("Clustering with " + best.size() + " + 1 centroids missing.");
                    List<Clusterable> more_centers = best.centers();
                    Clusterable to_split = more_centers.remove((new Random()).nextInt(more_centers.size()));
                    more_centers.addAll(to_split.children(best.dataForCenter(to_split)));
                    Segmentation segmentation = kMeans(more_centers, m_data_points, 0);
                    if (segmentation == null || segmentation.isEmpty())
                    {
                        System.err.println("FilTracker Clusterer.cluster: Global clustering failed.");
                        return null;
                    }
                    // This should work, because cluster_segmentation is defined anew on each run.
                    m_segmentations.add(segmentation);
                }

                if (lower_neighbor && upper_neighbor)
                {
                    live = false;
                }
            }

            return best.toMap();
        }
        if (type == SILHOUETTE)
        {
            // Find greatest ratio of silhouette decrease.
            long max_score = Long.MIN_VALUE;
            Segmentation best = null;
            for (int i = 0; i < m_segmentations.size() - 1; i++)
            {
                Segmentation test = m_segmentations.get(i);
                long test_score = test.m_score;
                if (test_score > Math.max(Long.MIN_VALUE, max_score))
                {
                    best = test;
                    max_score = test_score;
                }
            }
            if (best == null)
            {
                return null;
            }
            return best.toMap();
        }
        return m_segmentations.get(m_segmentations.size() - 1).toMap();
    }

    /**
     * Simply cluster for a fixed number of centers.
     * @param centers   A list of initial centroids.
     * @param number    The number of centroids with which to cluster.
     * @param precision Maximal distance a centroid may move to be considered unmoved.
     * @return          A map of centroids and their corresponding clusters.
     */
    public Map<Clusterable, List<Clusterable>> cluster(List<Clusterable> centers, int number,
                                                       long precision)
    {
        if (centers == null || centers.isEmpty())
        {
            System.err.println("FilTracker Clusterer.cluster: " +
                               "No valid initial center given.");
            return null;
        }
        while (centers.size() < number)
        {
            Clusterable to_split = centers.remove((new Random()).nextInt(centers.size()));
            centers.addAll(to_split.children(m_data_points));
        }
        while (centers.size() > number)
        {
            centers.remove((new Random()).nextInt(centers.size()));
        }

        // K-Means over all centers.
        Segmentation cluster_segmentation = kMeans(centers, m_data_points, precision);
        if (cluster_segmentation == null || cluster_segmentation.isEmpty())
        {
            System.err.println("FilTracker Clusterer.cluster: Global clustering failed.");
            return null;
        }
        m_segmentations.add(cluster_segmentation);
        return cluster_segmentation.toMap();
    }

    /**
     * Simply cluster for all numbers of centers from min to max.
     * @param centers   A list of initial centroids.
     * @param min       The minimum number of centroids at which the algorithm is started.
     * @param max       The maximum number of centroids at which the algorithm is stopped.
     * @param precision Maximal distance a centroid may move to be considered unmoved.
     * @return          A list of maps of centroids and their corresponding clusters.
     */
    public List<Map<Clusterable, List<Clusterable>>> getAll(List<Clusterable> centers,
                                                            int min, int max,
                                                            long precision)
           throws NotImplementedException
    {
        if (centers == null || centers.isEmpty())
        {
            System.err.println("FilTracker Clusterer.cluster: " +
                               "No valid initial center given.");
            return null;
        }

        min = (min > max ? max : min);
        while (centers.size() < min)
        {
            Clusterable to_split = centers.remove((new Random()).nextInt(centers.size()));
            centers.addAll(to_split.children(m_data_points));
        }
        boolean live = true;

        while (live)
        {
            // K-Means over all centers.
            Segmentation cluster_segmentation = kMeans(centers, m_data_points, precision);
            if (cluster_segmentation == null || cluster_segmentation.isEmpty())
            {
                System.err.println("FilTracker Clusterer.cluster: Global clustering failed.");
                return null;
            }
            // This should work, because cluster_segmentation is defined anew on each run.
            m_segmentations.add(cluster_segmentation);

            // Break if maximum number of centers is reached.
            if (cluster_segmentation.size() >= max)
            {
                live = false;
            }

            centers = cluster_segmentation.centers();
            Clusterable to_split = centers.remove((new Random()).nextInt(centers.size()));
            centers.addAll(to_split.children(cluster_segmentation.dataForCenter(to_split)));
        }
        List<Map<Clusterable, List<Clusterable>>> output = new ArrayList<>();
        for (Segmentation s : m_segmentations)
        {
            output.add(s.toMap());
        }
        return output;
    }

    /**
     * Get the center of the data set.
     * @param precision Maximal distance a centroid may move to be considered unmoved.
     * @return          A Clusterable representing the center of the data set.
     */
    Clusterable center(long precision)
    {
        List<Clusterable> centers = new ArrayList<>();
        centers.add(m_data_points.get((new Random()).nextInt(m_data_points.size())).clone());
        Map<Clusterable, List<Clusterable>> result = cluster(centers, 1, precision);
        if (result.size() != 1)
        {
            return null;
        }
        Clusterable[] center_array = new Clusterable[1];
        result.keySet().toArray(center_array);
        return center_array[0];
    }

    double sumOfSquares(Clusterable center)
    {
        if (m_segmentations == null)
        {
            return -1;
        }
        for (Segmentation segmentation : m_segmentations)
        {
            double tmp = segmentation.sumSquaresForCenter(center);
            if (tmp > 0)
            {
                return tmp;
            }
        }
        return -1;
    }

    private static Segmentation kMeans(List<Clusterable> old_centers,
                                       List<Clusterable> data_points, long precision)
    {
        if (old_centers == null || old_centers.isEmpty())
        {
            return null;
        }

        System.out.print("Starting k-Means with " + old_centers.size() + " centroids and " +
                           data_points.size() + " data points... ");
        precision *= precision;
        boolean moved = true;
        long time = System.nanoTime();
        int iter = 0;
        Segmentation cluster_segmentation = null;

        while (moved)
        {
            iter++;
            cluster_segmentation = new Segmentation();
            List<Clusterable> new_centers = new ArrayList<>();
            for (Clusterable center : old_centers)
            {
                new_centers.add(center.clone());
            }

            // Sort all data points to the respective closest clusters.
            for (Clusterable event : data_points)
            {
                long minimum_distance_square = Long.MAX_VALUE;
                Clusterable nearest = null;
                for (Clusterable center : old_centers)
                {
                    long distance_square = event.distanceSquare(center);
                    if (distance_square < minimum_distance_square)
                    {
                        minimum_distance_square = distance_square;
                        nearest = center;
                    }
                }
                if (nearest == null)
                {
                    System.err.println("FilTracker Clusterer.kMeans: " +
                    		           "Data point could not be assigned to center!");
                    return null;
                }
                cluster_segmentation.addToCluster(nearest, event, minimum_distance_square);
            }

            // Move centers to new locations and check if any center moved.
            moved = false;
            for (Clusterable center : new_centers)
            {
                List<Clusterable> data = cluster_segmentation.dataForCenter(center);
                if (data == null)
                {
                    System.err.println("Center has no points: " + center + "\nStep: " + iter);
                    moved = true;
                    data = new ArrayList<>();
                    data.add(data_points.get((new Random()).nextInt(data_points.size())).clone());
                }
                center = center.moveToCenter(data);
                boolean unmoved = false;
                if (precision == 0)
                {
                    for (Clusterable old_center : old_centers)
                    {
                        unmoved = unmoved || center.equals(old_center);
                    }
                }
                else
                {
                    for (Clusterable old_center : old_centers)
                    {
                        unmoved = unmoved || center.distanceSquare(old_center) < precision;
                    }
                }
                moved = moved || !unmoved;
            }

            old_centers = new ArrayList<>();
            for (Clusterable center : new_centers)
            {
                old_centers.add(center.clone());
            }
            if (iter % 100 == 0)
            {
                System.err.println("Iteration " + iter + ". Doubling precision threshold " + precision);
                precision *= 2;
            }
        }
        System.out.println("done after " + iter + " iterations in " +
                           Math.round((System.nanoTime() - time) / MF) / 1000.0 +
                           " seconds.");
        return cluster_segmentation;
    }

    private static boolean split_test (Segmentation parent, Segmentation children, int type)
                   throws NotImplementedException
    {
        if (children == null || children.size() != 2)
        {
            System.err.println("FilTracker Clusterer.cluster: Local clustering failed.");
            return false;
        }
        switch (type)
        {
        case BIC:
            return scoreBIC(children) > scoreBIC(parent);
        case GAUSS:
            return gaussTest(children.centers(), parent.data());
        case MIN_DIST:
            return parent == null;
        default:
            return false;
        }
    }

    /**
     * Score a cluster segmentation with respect to Kass-Wasserman form of
     * Bayesian Information Criterion.
     */
    private static long scoreBIC (Segmentation segmentation)
    {
        final int n_clusters = segmentation.size();
        if (n_clusters <= 0)
        {
            return Long.MIN_VALUE;
        }
        int[] cluster_sizes = segmentation.dataSize();
        int dimension = segmentation.dimension();
        double variance_MLLE = Math.log(segmentation.sumSquareDistances() / (MF * MF));
        double score = 0;
        for (int i = 0; i < n_clusters; i++)
        {
            score += cluster_sizes[i] * Math.log(cluster_sizes[i]);
        }
        score -= 0.5 * (n_clusters * (dimension + 1) * Math.log(cluster_sizes[n_clusters]) +
                        cluster_sizes[n_clusters] * dimension * variance_MLLE);
        return Math.round(score * MF);
    }

    /**
     * Score a cluster split with respect to the Anderson-Darling test.
     */
    private static boolean gaussTest(List<Clusterable> split, List<Clusterable> cluster)
                   throws NotImplementedException
    {
        final double threshold = 1.87;
        if (split == null || split.size() != 2 || cluster == null || cluster.isEmpty())
        {
            return false;
        }
        double[] projections = null;
        try
        {
            projections = split.get(0).projection(split, cluster);
        }
        catch(NotImplementedException e)
        {
            throw NotImplementedException.called_by(e, "Gaussian shape test");
        }

        Arrays.sort(projections);
        final int n = projections.length;
        final double root2 = Math.sqrt(2);
        for (int i = 0; i < projections.length; i++)
        {
            double tmp = 0.5 * (1 + Erf.erf(projections[i] / root2));
            projections[i] = tmp;
        }

        double score = -n * n;
        for (int i = 0; i < n; i++)
        {
            score -= (2 * i + 1) * (Math.log(projections[i]) + Math.log(1 - projections[n - i - 1]));
        }
        score *= (1 + (4 - (25 / (double)n)) / (double)n) / (double)n;

        System.out.println("Score: " + score);
        return score > threshold;
    }

    private long calculateSilhouette(Segmentation segmentation, long[][] matrix)
    {
        if (matrix == null)
        {
            // TODO
            return Long.MIN_VALUE;
        }
        List<Clusterable> data = segmentation.data();
        int[][] indices = getPositions(segmentation);
        double silhouette = 0;
        System.out.print("Start calculating silhouette... ");
        int counter = 0;
        for (Clusterable item : data)
        {
            counter++;
            if (counter % 1000 == 0)
            {
                System.out.print(counter + " ");
            }
            int item_index = m_data_points.indexOf(item);
            double mean_own = 0,
                   mean_nearest = Double.MAX_VALUE;
            for (int[] cluster_indices : indices)
            {
                double tmp = 0;
                boolean own = false;
                for (int that_index : cluster_indices)
                {
                    if (item_index > that_index)
                    {
                        tmp += matrix[item_index][that_index];
                    }
                    else
                    {
                        tmp += matrix[that_index][item_index];
                    }
                    if (item_index == that_index)
                    {
                        own = true;
                    }
                }
                if (own)
                {
                    mean_own = tmp / (cluster_indices.length - 1);
                }
                else
                {
                    tmp /= cluster_indices.length;
                    if (tmp < mean_nearest)
                    {
                        mean_nearest = tmp;
                    }
                }
                    
            }
            silhouette += (mean_nearest - mean_own) / Math.max(mean_own, mean_nearest);
        }
        return Math.round(silhouette * MF / (double)data.size());
    }

    private int[][] getPositions(Segmentation segmentation)
    {
        int[][] out = new int[segmentation.size()][];
        for (int i = 0; i < segmentation.size(); i++)
        {
            Cluster cluster = segmentation.m_clusters.get(i);
            out[i] = new int[cluster.size()];
            for (int j = 0; j < cluster.size(); j++)
            {
                out[i][j] = m_data_points.indexOf(cluster.m_data.get(j));
            }
        }
        return out;
    }

    private static class Cluster
    {
        private Clusterable m_center;
        private List<Clusterable> m_data;
        private double m_sum_distance_squares;

        private Cluster(Clusterable center, List<Clusterable> cluster)
        {
            m_center = center;
            addList(cluster);
        }

        private Cluster(Clusterable center, Clusterable item, double distance_square)
        {
            m_sum_distance_squares = distance_square;
            m_data = new ArrayList<>();
            m_data.add(item);
            m_center = center;
        }

        private boolean add(Clusterable item, long distance_square)
        {
            if (item == null || m_data == null || m_data.isEmpty())
            {
                return false;
            }
            m_sum_distance_squares += distance_square;
            return m_data.add(item);
        }

        private boolean add(Clusterable item)
        {
            if (item == null)
            {
                return false;
            }
            if (m_data == null)
            {
                m_data = new ArrayList<>();
            }
            m_sum_distance_squares += m_center.distanceSquare(item);
            return m_data.add(item);
        }

        private boolean addList(List<Clusterable> list)
        {
            if (list == null)
            {
                return false;
            }
            boolean out = true;
            for (Clusterable item : list)
            {
                out = out && add(item);
            }
            return out;
        }

        private int size()
        {
            if (m_data == null || m_data.isEmpty())
            {
                return 0;
            }
            return m_data.size();
        }
    }

    private static class Segmentation
    {
        private List<Cluster> m_clusters;
        private long m_score = Long.MIN_VALUE;

        private Segmentation (Cluster cluster)
        {
            if (cluster != null)
            {
                m_clusters = new ArrayList<>();
                m_clusters.add(cluster);
            }
        }

        private Segmentation ()
        {
            this(null);
        }

        private boolean isEmpty()
        {
            return m_clusters == null || m_clusters.isEmpty();
        }

        private int size()
        {
            if (m_clusters == null || m_clusters.isEmpty())
            {
                return 0;
            }
            return m_clusters.size();
        }

        private List<Clusterable> centers()
        {
            if (m_clusters == null || m_clusters.isEmpty())
            {
                return null;
            }
            List<Clusterable> centers = new ArrayList<>();
            for (Cluster cluster : m_clusters)
            {
                centers.add(cluster.m_center.clone());
            }
            return centers;
        }

        private List<Clusterable> data()
        {
            if (m_clusters == null || m_clusters.isEmpty())
            {
                return null;
            }
            if (m_clusters.size() == 1)
            {
                return m_clusters.get(0).m_data;
            }
            List<Clusterable> data = new ArrayList<>();
            for (Cluster cluster : m_clusters)
            {
                data.addAll(cluster.m_data);
            }
            return data;
        }

        private int[] dataSize()
        {
            if (m_clusters == null || m_clusters.isEmpty())
            {
                return null;
            }
            final int size = m_clusters.size();
            int[] out = new int[size + 1];
            for (int i = 0; i < size; i++)
            {
                out[i] = m_clusters.get(i).m_data.size();
                out[size] += out[i];
            }
            return out;
        }

        private int dimension()
        {
            if (m_clusters == null || m_clusters.isEmpty())
            {
                return -1;
            }
            return m_clusters.get(0).m_center.dimension();
        }

        private double sumSquareDistances ()
        {
            if (m_clusters == null || m_clusters.isEmpty())
            {
                return 0;
            }
            double out = 0;
            for (Cluster cluster : m_clusters)
            {
                out += cluster.m_sum_distance_squares;
            }
            return out;
        }

        private void addToCluster (Clusterable center, Clusterable item, long square_distance)
        {
            if (m_clusters == null)
            {
                m_clusters = new ArrayList<>();
            }
            boolean found = false;
            for (Cluster cluster : m_clusters)
            {
                if (center.equals(cluster.m_center))
                {
                    cluster.add(item, square_distance);
                    found = true;
                }
            }
            if (!found)
            {
                Cluster new_cluster = new Cluster(center, item, square_distance);
                m_clusters.add(new_cluster);
            }
        }

        private List<Clusterable> dataForCenter (Clusterable center)
        {
            if (m_clusters == null)
            {
                return null;
            }
            for (Cluster cluster : m_clusters)
            {
                if (center.equals(cluster.m_center))
                {
                    return cluster.m_data;
                }
            }
            return null;
        }

        private double sumSquaresForCenter (Clusterable center)
        {
            if (m_clusters == null)
            {
                return -1;
            }
            for (Cluster cluster : m_clusters)
            {
                if (center.equals(cluster.m_center))
                {
                    return cluster.m_sum_distance_squares;
                }
            }
            return -1;
        }

        private Map<Clusterable, List<Clusterable>> toMap ()
        {
            if (m_clusters == null)
            {
                return null;
            }
            Map<Clusterable, List<Clusterable>> out = new HashMap<>();
            for (Cluster cluster : m_clusters)
            {
                out.put(cluster.m_center, cluster.m_data);
            }
            return out;
        }
    }
}
