package clustering;
/*
 * This file is part of CANE - A Clustering Algorithm for Non-Euclidean spaces.
 *
 * Copyright (C) 2014 Benjamin Eltzner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Random;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;


/**
 * This class constitutes a hierarchy of clusters. The hierarchy
 * is represented in terms of a binary tree where the first branch
 * at each node represents the largest cluster and the second
 * branch represents the rest of the data.
 * 
 * The tree can be accessed from the root or from the leaves and
 * can be walked using the three branch functions of the nodes.
 * Data points can be retrieved from the leaves.
 */
public class TreeClusterer implements HierarchicalClusterer
{

    private List<Node> m_nodes;
    private Node m_root;

    public TreeClusterer (List<Clusterable> data)
    {
        Node root = new Node(null, null, data);
        m_root = root;
        m_nodes = new ArrayList<>();
        m_nodes.add(root);
    }

    public void cluster(int branches, int limit, long precision)
    {
        while (true)
        {
            List<Node> new_leaves = new ArrayList<>();
            for (Node n : m_nodes)
            {
                if (n.size() > limit)
                {
                    new_leaves.addAll(n.fork(branches, precision));
                }
            }
            m_nodes.addAll(new_leaves);
            if (new_leaves.isEmpty())
            {
                break;
            }
        }
    }

    /* (non-Javadoc)
     * @see clustering.HierarchicalClusterer#root()
     */
    @Override
    public Node root()
    {
        return m_root;
    }

    /* (non-Javadoc)
     * @see clustering.HierarchicalClusterer#leaves()
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Node> leaves()
    {
        List<Node> leaves = new ArrayList<>();
        for (Node n : m_nodes)
        {
            if (n.isLeaf())
            {
                leaves.add(n);
            }
        }
        return leaves;
    }

    /* (non-Javadoc)
     * @see clustering.HierarchicalClusterer#toXML()
     */
    @Override
    public Document toXML()
    {
        Document doc = null;
        try
        {
            // Initialize documentBuilder
            doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
        }
        catch (ParserConfigurationException e)
        {
            System.out.println("Fatal: failed to create DocumentBuilder! " + e.getMessage());
            return null;
        }
        return walk(m_root, null, doc, "");
    }

    /**
     * Adds the hardest cluster to the segmentation and returns the remaining data points.
     */
    public List<Clusterable> biggestCluster (Map<Clusterable, List<Clusterable>> segmentation, double min)
    {
        Node biggest = null;
        int max = 0;
        for (Node node : m_nodes)
        {
            if (node.m_hardness > min && node.m_size > max)
            {
                max = node.m_size;
                biggest = node;
            }
        }
        if (biggest == null)
        {
            return m_root.getPoints();
        }
        List<Clusterable> cluster = biggest.getPoints();
        segmentation.put(biggest.m_self, cluster);
        List<Clusterable> out = m_root.getPoints();
        out.removeAll(cluster);
        return out;
    }

    private static Document walk(Node node, Element parent, Document doc, String label)
    {
        if (node == null)
        {
            return null;
        }
        Element element = node.element(doc, label);
        if (parent == null)
        {
            doc.appendChild(element);
        }
        else
        {
            parent.appendChild(element);
        }
        walk(node.child1(), element, doc, label + "0");
        walk(node.child2(), element, doc, label + "1");
        return doc;
    }

    public static class Node extends AbstractNode<Node>
    {
        final private Clusterable m_self;
        private List<Clusterable> m_data;
        final private int m_size;
        private double m_hardness;

        private Node(Node parent, Clusterable self, double squares_sum, List<Clusterable> data)
        {
            m_hardness = 0;
            setParent(parent);
            if (self == null && data != null)
            {
                m_self = (new Clusterer(data)).center(0);
                m_self.exportCenter();
            }
            else
            {
                m_self = self;
            }
            m_data = data;
            if (m_self != null && m_data != null)
            {
                m_size = data.size();
            }
            else
            {
                m_size = -1;
            }
        }

        private Node(Node parent, Clusterable self, List<Clusterable> data)
        {
            m_hardness = 0;
            setParent(parent);
            double squares_sum = -1;
            if (self == null && data != null)
            {
                Clusterer clusterer = new Clusterer(data); 
                m_self = clusterer.center(0);
                squares_sum = clusterer.sumOfSquares(m_self);
            }
            else
            {
                m_self = self;
            }
            m_data = data;
            if (squares_sum > 0 && m_self != null && m_data != null)
            {
                m_size = data.size();
            }
            else
            {
                m_size = -1;
            }
        }

        private List<Node> fork(int branches, long precision)
        {
            if (m_data == null)
            {
                return null;
            }

            Clusterer clusterer = new Clusterer(m_data);
            List<Clusterable> centers = new ArrayList<>();
            if (m_self == null)
            {
                centers.add(m_data.get((new Random()).nextInt(m_data.size())).clone());
            }
            else
            {
                centers.add(m_self);
            }
            Map<Clusterable, List<Clusterable>> cluster_result = 
                    clusterer.cluster(centers, branches, precision);

            int max = 0;
            Clusterable main_child = null;
            List<Clusterable> main_data = new ArrayList<>(),
                              other_data = new ArrayList<>();
            for (Clusterable center : cluster_result.keySet())
            {
                List<Clusterable> data = cluster_result.get(center);
                if (data.size() > max)
                {
                    max = data.size();
                    other_data.addAll(main_data);
                    main_child = center;
                    main_data = data;
                }
                else
                {
                    other_data.addAll(data);
                }
            }

            double sum_of_squares = clusterer.sumOfSquares(main_child);
            main_child.exportCenter();
            Node child1 = new Node(this, main_child, sum_of_squares, main_data);
            setChild1(child1);
            if (parent() != null)
            {
                m_hardness = m_size * m_size / ((float)(parent().m_size * child1.m_size)); 
            }
            Node child2 = new Node(this, null, other_data);
            setChild2(child2);
            m_data = null;
            List<Node> out = new ArrayList<>();
            out.add(child1);
            out.add(child2);
            return out;
        }

        public boolean isLeaf()
        {
            return m_data != null;
        }

        /**
         * Returns a deep copy of the data in a leaf to protect
         * integrity of the tree.
         */
        public List<Clusterable> data()
        {
            if (m_data == null)
            {
                return null;
            }
            List<Clusterable> out = new ArrayList<>();
            for (Clusterable p : m_data)
            {
                out.add(p.clone());
            }
            return out;
        }

        public List<Clusterable> getPoints()
        {
            if (isLeaf())
            {
                return data();
            }

            List<Clusterable> out = child1().getPoints();
            out.addAll(child2().getPoints());
            return out;
        }

        @Override
        public String toString()
        {
            String out = "";
            if (m_self == null)
            {
                out += "Rest ";
            }
            else
            {
               out += m_self.toString() + " ";
            }
            if (isLeaf())
            {
                out += "leaf with " + size() + " data points";
            }
            else
            {
                out += "crotch";
            }
            return out;
        }

        private int size()
        {
            if (m_data == null)
            {
                return 0;
            }
            return m_data.size();
        }

        private Element element(Document doc, String identifier)
        {
            Element this_element = doc.createElement("center" + identifier);
            this_element.setAttribute("position", m_self.exportCenter());
            this_element.setAttribute("hardness", "" + Math.round(m_hardness * 1.0e6) * 1.0e-6);
            this_element.setAttribute("size", "" + m_size);
            if (isLeaf())
            {
                int i = 0;
                for (Clusterable item : m_data)
                {
                    this_element.setAttribute("point" + i++, "" + item.label());
                }
            }
            return this_element;
        }
    }
}
