package main;
/*
 * This file is part of CANE - A Clustering Algorithm for Non-Euclidean spaces.
 *
 * Copyright (C) 2014 Benjamin Eltzner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import types.ImportData;
import clustering.Clusterable;
import clustering.Clusterer;
import clustering.SLINKClusterer;
import clustering.TreeClusterer;

class MainDialog
{

    private static final int SINGLE = 0, LOOP = 1;

    public MainDialog()
    {
        switch(loopDialog("Do a single run or an automated RESH run?"))
        {
        case SINGLE:
            singleRunDialog();
            break;
        case LOOP:
            loopRunDialog();
            break;
        default:
            System.out.println("Operation aborted. Have a nice day!");
            return;
        }
        System.out.println("Operation finished. Have a nice day!");
    }

    private static void singleRun(List<Clusterable> data, File data_file, int type)
    {
        Clusterer clusterer = new Clusterer(data);
        Random random = new Random();
        List<Clusterable> centers = new ArrayList<>();
        centers.add(data.get(random.nextInt(data.size())).clone());
        long time = -System.nanoTime();
        List<Map<Clusterable, List<Clusterable>>> result = clusterer.getAll(centers, 3, 3, 10000);
        time += System.nanoTime();
        System.out.println("Done after " + Math.round(time / (double)1e6) / 1000.0 + " seconds.");
        for (Map<Clusterable, List<Clusterable>> segmentation : result)
        {
            System.out.println("Best fit for " + segmentation.size() + " clusters:");
            for (Clusterable center : segmentation.keySet())
            {
                System.out.println(center.exportCenter() + " " + segmentation.get(center).size());
            }
            try
            {
                String tmp = data_file.getCanonicalPath().replaceAll("\\.[A-Za-z0-9]{3,4}$", "") +
                             "_" + segmentation.size() + "_" + ImportData.type(type) +
                             "_cluster_result.csv";
                File export_file = new File(tmp);
                ImportData.exportCSV(export_file, segmentation);
            }
            catch (IOException e)
            {
                System.err.println("Could not save data: " + e.getMessage());
            }
        }
    }

    private static void slink(List<Clusterable> data, File data_file, int type)
    {
        SLINKClusterer tree = new SLINKClusterer(data);
        long time = -System.nanoTime();
        tree.cluster();
        time += System.nanoTime();
        System.out.println("Done after " + Math.round(time / (double)1e6) / 1000.0 + " seconds.");
        File export_file = null;

        try
        {
            String tmp = data_file.getCanonicalPath().replaceAll("\\.[A-Za-z0-9]{3,4}$", "") +
                                                                 "_" + ImportData.type(type) +
                                                                 "_SLINK_result.xml";
            export_file = new File(tmp);
            // Write xml file
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");

            FileWriter writer = new FileWriter(export_file);
            StreamResult result = new StreamResult(writer);
            DOMSource source = new DOMSource(tree.toXML());
            transformer.transform(source, result);
            writer.close();
        }
        catch (Exception e) // IOException or TransformerConfigurationException
        {
            System.err.println("Error while saving '" + export_file + "': " + e.getMessage());
            return;
        }
    }
 
    private static void multiSlink(List<Clusterable> data, File data_file, int type)
    {
        SLINKClusterer tree = new SLINKClusterer(data);
        long time = -System.nanoTime();
        List<List<Clusterable>> segmentation = tree.greatestClusters(15, 50 * 1000000);
        time += System.nanoTime();
        System.out.println("Done after " + Math.round(time / (double)1e6) / 1000.0 + " seconds.");
        System.out.println("Best fit for " + segmentation.size() + " clusters:");
        for (List<Clusterable> cluster : segmentation)
        {
            System.out.println(cluster.size() + " points.");
        }
        
        File export_file = null;
        try
        {
            String tmp = data_file.getCanonicalPath().replaceAll("\\.[A-Za-z0-9]{3,4}$", "") +
                         "_" + segmentation.size() + "_" + ImportData.type(type) +
                         "_multiSLINK_result.csv";
            export_file = new File(tmp);
            ImportData.exportCSV(export_file, segmentation);
        }
        catch (IOException e)
        {
            System.err.println("Error while saving '" + export_file + "': " + e.getMessage());
        }
    }

    private static void singleTree(List<Clusterable> data, File data_file, int type)
    {
        TreeClusterer tree = new TreeClusterer(data);
        long time = -System.nanoTime();
        tree.cluster(3, 100, 2000);
        time += System.nanoTime();
        System.out.println("Done after " + Math.round(time / (double)1e6) / 1000.0 + " seconds.");
        File export_file = null;

        try
        {
            String tmp = data_file.getCanonicalPath().replaceAll("\\.[A-Za-z0-9]{3,4}$", "") +
                                                                 "_" + ImportData.type(type) +
                                                                 "_tree_result.xml";
            export_file = new File(tmp);
            // Write xml file
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");

            FileWriter writer = new FileWriter(export_file);
            StreamResult result = new StreamResult(writer);
            DOMSource source = new DOMSource(tree.toXML());
            transformer.transform(source, result);
            writer.close();
        }
        catch (Exception e) // IOException or TransformerConfigurationException
        {
            System.err.println("Error while saving '" + export_file + "': " + e.getMessage());
            return;
        }
    }

    private static void forest(List<Clusterable> data, File data_file, int type)
    {
        Map<Clusterable, List<Clusterable>> segmentation = new HashMap<>();
        int old_size = data.size() + 1;
        while (data.size() < old_size)
        {
            old_size = data.size();
            TreeClusterer tree = new TreeClusterer(data);
            Random random = new Random();
            List<Clusterable> centers = new ArrayList<>();
            centers.add(data.get(random.nextInt(data.size())).clone());
            long time = -System.nanoTime();
            tree.cluster(2, Math.max(data.size() / 10, 10), 0);
            data = tree.biggestCluster(segmentation, 1.3);
            time += System.nanoTime();
            System.out.println("Done after " + Math.round(time / (double)1e6) / 1000.0 + " seconds.");
        }
        System.out.println("Best fit for " + segmentation.size() + " clusters:");
        for (Clusterable center : segmentation.keySet())
        {
            System.out.println(center.exportCenter() + " " + segmentation.get(center).size());
        }
        try
        {
            String tmp = data_file.getCanonicalPath().replaceAll("\\.[A-Za-z0-9]{3,4}$", "") +
                         "_" + segmentation.size() + "_" + ImportData.type(type) +
                         "_forest_cluster_result.csv";
            File export_file = new File(tmp);
            ImportData.exportCSV(export_file, segmentation);
        }
        catch (IOException e)
        {
            System.err.println("Could not save data: " + e.getMessage());
        }
    }

    private static void singleRunDialog()
    {
        boolean loop = true;
        File data_file = null;
        List<Clusterable> data = null;
        int type = -1;
        while (loop)
        {
            type = modeDialog("Choose data type!");
            if (type == -1)
            {
                System.out.println("Operation aborted. Have a nice day!");
                return;
            }
            data_file = loadDialog("Choose data file.", false);
            data = ImportData.importCSV(type, data_file);
            if (data == null || data.isEmpty())
            {
                if (JOptionPane.showOptionDialog(null, "Importing data failed. Try again?",
                        "Retry importing data?", JOptionPane.YES_NO_OPTION,
                        JOptionPane.QUESTION_MESSAGE, null, null, null) != JOptionPane.YES_OPTION)
                {
                    System.out.println("Operation aborted. Have a nice day!");
                    return;
                }
            }
            else
            {
                loop = false;
            }
        }

        multiSlink(data, data_file, type);
    }

    private static void loopRunDialog()
    {
        int type = modeDialog("Choose data type!");
        if (type == -1)
        {
            System.out.println("Operation aborted. Have a nice day!");
            return;
        }

        Path path = loadDialog("Choose data path.", true).toPath();
        final List<File> file_list = new ArrayList<>();
        try
        {
            Files.walkFileTree(path, new SimpleFileVisitor<Path>() {
                @Override
                public FileVisitResult visitFile(Path file, BasicFileAttributes attrs)
                                       throws IOException
                {
                    String file_name = file.getFileName().toString();
                    if (file_name.endsWith(".csv") &&
                        file_name.contains("RESH") &&
                        !file_name.contains("result"))
                    {
                        file_list.add(file.toFile());
                    }
                    return FileVisitResult.CONTINUE;
                }
            });

            for (File file : file_list)
            {
                System.out.println(file.getCanonicalPath());
                List<Clusterable> data = ImportData.importCSV(type, file);
                if (data == null || data.isEmpty())
                {
                    System.err.println("Could not read data!");
                }
                else
                {
                    singleRun(data, file, type);
                }
            }
        }
        catch (IOException e)
        {
            System.err.println("IOException: " + e.getMessage());
        }
    }

    private static int loopDialog(String message)
    {
        Map<String, Integer> map = new HashMap<String, Integer>();
        map.put("Single run", SINGLE);
        map.put("Automated RESH loop", LOOP);

        Object[] options = map.keySet().toArray();
        Object choice = JOptionPane.showInputDialog(null, message, "Data Type",
                            JOptionPane.QUESTION_MESSAGE, null, options, options[0]);
        if (choice == null)
        {
            return -1;
        }
        String mode = (String) choice;
        return map.get(mode);
    }

    private static int modeDialog(String message)
    {
        Map<String, Integer> map = ImportData.modes();

        Object[] options = map.keySet().toArray();
        Object choice = JOptionPane.showInputDialog(null, message, "Data Type",
                            JOptionPane.QUESTION_MESSAGE, null, options, options[0]);
        if (choice == null)
        {
            return -1;
        }
        String mode = (String) choice;
        return map.get(mode);
    }

    private static File loadDialog(String caption, boolean folder)
    {
        JFileChooser fc = new JFileChooser();
        fc.setDialogTitle(caption);
        fc.setDialogType(JFileChooser.OPEN_DIALOG);
        fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
        fc.setAcceptAllFileFilterUsed(false);
        String path = System.getProperty("user.home");

        File pathFile = new File(path);
        fc.setCurrentDirectory(pathFile);

        if (folder)
        {
            fc.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
            if (fc.showOpenDialog(null) == JFileChooser.APPROVE_OPTION)
            {
                File file = fc.getSelectedFile();

                if (file == null)
                {
                    return null;
                }

                if (file.isDirectory())
                {
                    return fc.getSelectedFile();
                }
                return fc.getSelectedFile().getParentFile();
            }
            return null;
        }

        if (fc.showOpenDialog(null) == JFileChooser.APPROVE_OPTION)
        {
            if (fc.getSelectedFile() == null)
            {
                return null;
            }

            return fc.getSelectedFile();
        }
        return null;
    }

}
