package types;
/*
 * This file is part of CANE - A Clustering Algorithm for Non-Euclidean spaces.
 *
 * Copyright (C) 2014 Benjamin Eltzner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.util.Arrays;

abstract class Calc
{

    public static long[] arrayRound(double[] in)
    {
        if (in == null)
        {
            return null;
        }
        long[] out = new long[in.length];
        for (int i = 0; i < in.length; i++)
        {
            out[i] = Math.round(in[i]);
        }
        return out;
    }

    public static double[] l2d(long[] in)
    {
        if (in == null)
        {
            return null;
        }
        double[] out = new double[in.length];
        for (int i = 0; i < in.length; i++)
        {
            out[i] = in[i];
        }
        return out;
    }

    public static double circularMean(double[] angles, double pi)
    {
        int n = angles.length;
        double mean0 = 0.0,
               var0 = 0.0;

        for (int i = 0; i < n; i++)
        {
            angles[i] = (angles[i] > pi ? angles[i] - 2 * pi : angles[i]);
            mean0 += angles[i];
            var0 += angles[i] * angles[i];
        }
        mean0 /= (double)n;
        var0 = var0 / (double)n - mean0 * mean0;
        Arrays.sort(angles);

        double[] means = new double[n];
        for (int i = 0; i < n; i++)
        {
            means[i] = mean0 + 2 * pi * i / (float) n;
            means[i] = (means[i] > pi ? means[i] - 2 * pi : means[i]);
        }

        double[] parts = new double[n];
        parts[0] = (means[0] > 0 ? means[0] : 0);
        for (int i = 0; i < n; i++)
        {
            if (means[i] < 0)
            {
                for (int j = i; j < n; j++)
                {
                    parts[i] += angles[j];
                }
                parts[i] /= (double)(n - i);
            }
            else
            {
                for (int j = 0; j < i; j++)
                {
                    parts[i] += angles[j];
                }
                parts[i] /= (double)i;
            }
        }

        double var_min = var0;
        int min_location = 0;
        for (int i = 1; i < n; i++)
        {
            double var = 0.0;
            if (means[i] < 0)
            {
                var = var0 + ((4 * pi * (n - i) / (double)n) * (pi - parts[i] + mean0) -
                              (2 * pi * (n - i) / (double)n) * (2 * pi * (n - i) / (double)n));
            }
            else
            {
                var = var0 + ((4 * pi * i / (double)n) * (pi + parts[i] - mean0) -
                              (2 * pi * i / (double)n) * (2 * pi * i / (double)n));
            }
            if (var < var_min)
            {
                var_min = var;
                min_location = i;
            }
        }

        return (2 * pi + means[min_location]) % (2 * pi);
    }

    public static long circularMean(long[] angles, long pi)
    {
        int n = angles.length;
        double mean0 = 0.0,
               var0 = 0.0;

        for (int i = 0; i < n; i++)
        {
            angles[i] = (angles[i] > pi ? angles[i] - 2 * pi : angles[i]);
            mean0 += angles[i];
            var0 += angles[i] * angles[i];
        }
        mean0 /= (double)n;
        var0 = var0 / (double)n - mean0 * mean0;
        Arrays.sort(angles);

        double[] means = new double[n];
        for (int i = 0; i < n; i++)
        {
            means[i] = mean0 + 2 * pi * i / (float) n;
            means[i] = (means[i] > pi ? means[i] - 2 * pi : means[i]);
        }

        double[] parts = new double[n];
        parts[0] = (means[0] > 0 ? means[0] : 0);
        for (int i = 0; i < n; i++)
        {
            if (means[i] < 0)
            {
                for (int j = i; j < n; j++)
                {
                    parts[i] += angles[j];
                }
                parts[i] /= (double)(n - i);
            }
            else
            {
                for (int j = 0; j < i; j++)
                {
                    parts[i] += angles[j];
                }
                parts[i] /= (double)i;
            }
        }

        double var_min = var0;
        int min_location = 0;
        for (int i = 1; i < n; i++)
        {
            double var = 0.0;
            if (means[i] < 0)
            {
                var = var0 + ((4 * pi * (n - i) / (double)n) * (pi - parts[i] + mean0) -
                              (2 * pi * (n - i) / (double)n) * (2 * pi * (n - i) / (double)n));
            }
            else
            {
                var = var0 + ((4 * pi * i / (double)n) * (pi + parts[i] - mean0) -
                              (2 * pi * i / (double)n) * (2 * pi * i / (double)n));
            }
            if (var < var_min)
            {
                var_min = var;
                min_location = i;
            }
        }

        return Math.round((2 * pi + means[min_location]) % (2 * pi));
    }

    public static long pow(int base, int exp)
    {
        long out = 1;
        for (int i = 0; i < exp; i++)
        {
            out *= base;
        }
        return out;
    }

    public static double[] projectRESH (double[] array)
    {
        final int l = array.length - 1;
        for (int i = 0; i < l; i++)
        {
            while (array[i] < 0)
            {
                array[i] += 180 * Const.MF;
            }
            array[i] = array[i] % (180 * Const.MF);
        }
        while (array[l] < 0)
        {
            array[l] += 180 * Const.MF;
        }
        array[l] = array[l] % (360 * Const.MF);
        return array;
    }

    public static long[] projectRESH (long[] array)
    {
        final int l = array.length - 1;
        for (int i = 0; i < l; i++)
        {
            while (array[i] < 0)
            {
                array[i] += 180 * Const.M;
            }
            array[i] = array[i] % (180 * Const.M);
        }
        while (array[l] < 0)
        {
            array[l] += 180 * Const.M;
        }
        array[l] = array[l] % (360 * Const.M);
        return array;
    }
}
