package types;
/*
 * This file is part of CANE - A Clustering Algorithm for Non-Euclidean spaces.
 *
 * Copyright (C) 2014 Benjamin Eltzner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.util.Arrays;
import java.util.List;

import clustering.Clusterable;
import clustering.NotImplementedException;
import clustering.SimpleClusterable;

/**
 * Let
 * n : The number of original angles
 * d : The dimension of the sphere
 * 
 * The constructor of this class corresponds to the representation of
 * points in the RESH space as vectors in R^(n+1) with length 10^6.
 * 
 * The set() and distance() methods expect input corresponding to the
 * representation of points as arrays of d-1 angles in the range [0, 180E+6)
 * and one angle in the range [0, 360E+6).
 */
class ClusterableSphere implements SimpleClusterable
{

    private long[] m_point;
    private long[] m_angles;
    private int m_label;

    public ClusterableSphere(long[] point, int label)
    {
        if (point.length < 8)
        {
            System.err.println("Strange point dimension: " + point.length);
        }
        double[] projected = Subsphere.project(Calc.l2d(point));
        m_point = Calc.arrayRound(projected);
        m_angles = Calc.arrayRound(Subsphere.v2a(projected));
        m_label = label;
    }

    /**
     * Constructor for the clone() method.
     */
    private ClusterableSphere(long[] point, long[] angles, int label)
    {
        m_point = Arrays.copyOf(point, point.length);
        m_angles = Arrays.copyOf(angles, angles.length);
        m_label = label;
    }

    @Override
    public String exportCenter()
    {
        long[] tmp = Calc.arrayRound(Subsphere.allAngles(Calc.l2d(m_angles)));
        return Arrays.toString(tmp).replace("[", "").replace(" ", "").replace("]", "");
    }

    @Override
    public int label() throws NotImplementedException
    {
        return m_label;
    }

    @Override
    public int dimension()
    {
        return m_angles.length;
    }

    @Override
    public double distance(double[] that) throws IllegalArgumentException
    {
        return distance_calculation(Subsphere.a2v(that));
    }

    @Override
    public double distance(SimpleClusterable that)
    {
        if (that == null || !that.getClass().equals(this.getClass()))
        {
            return Long.MAX_VALUE;
        }
        return distance_calculation(Calc.l2d(((ClusterableSphere) that).m_point));
    }

    private double distance_calculation(double[] that)
    {
        // Standard spherical distance calculation
        int length = that.length;
        if (length != m_point.length)
        {
            System.err.println("Incompatible lengths: " + m_point.length + " and " + length);
        }
        double dist = 0,
               this_norm_sq = 0,
               that_norm_sq = 0;
        for (int i = 0; i < length; i++)
        {
            dist += that[i] * m_point[i];
            this_norm_sq += m_point[i] * m_point[i];
            that_norm_sq += that[i] * that[i];
        }
        dist /= Math.sqrt(this_norm_sq * that_norm_sq);
        dist = (dist >  1 ? 0 :
               (dist < -1 ? 180 * Const.MF :
                            Const.MF * Math.acos(dist) / Const.RAD));

        return dist;
    }

    @Override
    public double[] projection(List<Clusterable> direction, List<Clusterable> cluster)
           throws NotImplementedException
    {
        throw NotImplementedException.throw_new("ClusterableSphere.projection()");
    }

    @Override
    public double[] toArray()
    {
        return Calc.l2d(m_angles);
    }

    /**
     * This method constructs the object from d (the dimension
     * of the RESH space) angles.
     */
    @Override
    public SimpleClusterable set(double[] that, boolean move_to_center)
    {
        that = Calc.projectRESH(that);
        m_angles = Calc.projectRESH(Calc.arrayRound(that));
        m_point = Calc.arrayRound(Subsphere.a2v(that));
        return this;
    }

    @Override
    public String toString()
    {
        return "ClusterableSphere [m_angles=" + Arrays.toString(m_angles) + "]";
    }

    public SimpleClusterable clone()
    {
        return new ClusterableSphere(m_point, m_angles, m_label);
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + Arrays.hashCode(m_angles);
        result = prime * result + m_label;
        result = prime * result + Arrays.hashCode(m_point);
        return result;
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        if (obj == null || getClass() != obj.getClass())
        {
            return false;
        }
        ClusterableSphere other = (ClusterableSphere) obj;
        if (!Arrays.equals(m_angles, other.m_angles) || m_label != other.m_label ||
            !Arrays.equals(m_point, other.m_point))
        {
            return false;
        }
        return true;
    }

}