package types;
/*
 * This file is part of CANE - A Clustering Algorithm for Non-Euclidean spaces.
 *
 * Copyright (C) 2015 Benjamin Eltzner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Random;

import clustering.Clusterable;
import clustering.NotImplementedException;

public class ClusterableTimeSeries implements Clusterable
{

    private long[] m_vector;
    private int m_label;

    public ClusterableTimeSeries(long[] vector, int label)
    {
        if (vector == null)
        {
            return;
        }
        m_vector = new long[vector.length];
        for (int i = 0; i < vector.length; i++)
        {
            m_vector[i] = vector[i];
        }
        m_label = label;
    }

    @Override
    public long distanceSquare(Clusterable that)
    {
        if (that == null || !that.getClass().equals(this.getClass()))
        {
            return Long.MAX_VALUE;
        }
        ClusterableTimeSeries other = (ClusterableTimeSeries) that;
        if (m_vector.length != other.m_vector.length)
        {
            return Long.MAX_VALUE;
        }

        long out = 0;
        for (int i = 0; i < m_vector.length; i++)
        {
            out += (m_vector[i] - other.m_vector[i]) * (m_vector[i] - other.m_vector[i]);
        }
        return out;
    }

    @Override
    public Clusterable moveToCenter(List<Clusterable> cluster)
    {
        if (cluster == null || cluster.isEmpty())
        {
            System.err.println("Cluster " + this + " has no points!");
            return setRandom();
        }
        long[] vector = new long[m_vector.length];
        for (Clusterable that : cluster)
        {
            if (!that.getClass().equals(this.getClass()))
            {
                return null;
            }
            ClusterableTimeSeries other = (ClusterableTimeSeries) that;
            if (m_vector.length != other.m_vector.length)
            {
                return null;
            }
            for (int i = 0; i < m_vector.length; i++)
            {
                vector[i] += other.m_vector[i];
            }
        }
        for (int i = 0; i < m_vector.length; i++)
        {
            m_vector[i] = vector[i] / cluster.size();
        }
        
        return this;
    }

    @Override
    public List<Clusterable> children(List<Clusterable> cluster)
    {
        Random random = new Random();
        int d = m_vector.length;
        double[] shift = new double[d];
        double shift_norm = 0;
        for (int i = 0; i < d; i++)
        {
            shift[i] = random.nextDouble();
            shift_norm += shift[i] * shift[i];
        }
        shift_norm = Math.sqrt(shift_norm / this.distanceSquare(cluster.get(random.nextInt(cluster.size()))));
        double[] shift_minus = new double[d];
        for (int i = 0; i < d; i++)
        {
            shift[i] /= shift_norm;
            shift_minus[i] = - shift[i];
        }

        ClusterableTimeSeries child1 = (ClusterableTimeSeries)this.clone();
        ClusterableTimeSeries child2 = (ClusterableTimeSeries)this.clone();
        List<Clusterable> out = new ArrayList<>();
        out.add(child1.move(shift));
        out.add(child2.move(shift_minus));
        return out;
    }

    @Override
    public double[] projection(List<Clusterable> direction, List<Clusterable> cluster) throws NotImplementedException
    {
        throw NotImplementedException.throw_new("ClusterableTimeSeries.projection()");
    }

    @Override
    public String exportCenter()
    {
        return Arrays.toString(m_vector).replace("[", "").replace(" ", "").replace("]", "");
    }

    @Override
    public int label() throws NotImplementedException
    {
        return m_label;
    }

    @Override
    public int dimension()
    {
        return m_vector.length;
    }

    private Clusterable move(double[] shift)
    {
        for (int i = 0; i < m_vector.length; i++)
        {
            m_vector[i] = Math.round(m_vector[i] + shift[i]);
        }
        return this;
    }

    private Clusterable setRandom()
    {
        Random random = new Random();
        int d = m_vector.length;
        for (int i = 0; i < d; i++)
        {
            m_vector[i] = Math.round(random.nextDouble() * 360.0 * Const.MF);
        }
        return this;
    }

    @Override
    public Clusterable clone()
    {
        return new ClusterableTimeSeries(m_vector, m_label);
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + Arrays.hashCode(m_vector);
        return result;
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        if (obj == null || getClass() != obj.getClass())
        {
            return false;
        }
        ClusterableTimeSeries other = (ClusterableTimeSeries) obj;
        if (m_label != other.m_label || !Arrays.equals(m_vector, other.m_vector))
        {
            return false;
        }
        return true;
    }

}
