package types;
/*
 * This file is part of CANE - A Clustering Algorithm for Non-Euclidean spaces.
 *
 * Copyright (C) 2014 Benjamin Eltzner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Random;

import clustering.Clusterable;


class ClusterableTorus implements Clusterable
{

    private long[] m_angles;
    private int m_label;
    private static final long PI = 180 * Const.M;
    private static final double EPS = 0.0000000001;

    public ClusterableTorus(long[] angles, int label)
    {
        if (angles == null)
        {
            return;
        }
        m_angles = new long[angles.length];
        for (int i = 0; i < angles.length; i++)
        {
            m_angles[i] = (angles[i] + 2 * PI) % (2 * PI);
        }
        m_label = label;
    }

    @Override
    public String exportCenter()
    {
        return Arrays.toString(m_angles).replace("[", "").replace(" ", "").replace("]", "");
    }

    @Override
    public int label()
    {
        return m_label;
    }

    @Override
    public int dimension()
    {
        return m_angles.length;
    }

    @Override
    public long distanceSquare(Clusterable that)
    {
        if (that == null || !that.getClass().equals(this.getClass()))
        {
            return Long.MAX_VALUE;
        }
        ClusterableTorus other = (ClusterableTorus) that;
        if (m_angles.length != other.m_angles.length)
        {
            return Long.MAX_VALUE;
        }

        long out = 0;
        for (int i = 0; i < m_angles.length; i++)
        {
            long tmp = Math.abs(m_angles[i] - other.m_angles[i]);
            out += (tmp > PI ? (2 * PI - tmp) * (2 * PI - tmp) : tmp * tmp);
        }
        return out;
    }

    @Override
    public Clusterable moveToCenter(List<Clusterable> cluster)
    {
        if (cluster == null || cluster.isEmpty())
        {
            System.err.println("Cluster " + this + " has no points!");
            return setRandom();
        }
        long[][] angles = new long[m_angles.length][cluster.size()];
        for (int n = 0; n < cluster.size(); n++)
        {
            Clusterable that = cluster.get(n);
            if (!that.getClass().equals(this.getClass()))
            {
                return null;
            }
            ClusterableTorus other = (ClusterableTorus) that;
            if (m_angles.length != other.m_angles.length)
            {
                return null;
            }
            for (int i = 0; i < m_angles.length; i++)
            {
                angles[i][n] = other.m_angles[i];
            }
        }
        for (int i = 0; i < m_angles.length; i++)
        {
            m_angles[i] = Calc.circularMean(angles[i], PI);
        }
        
        return this;
    }

    @Override
    public String toString()
    {
        return "ClusterableTorus [m_angles=" + Arrays.toString(m_angles) + "]";
    }

    @Override
    public List<Clusterable> children(List<Clusterable> cluster)
    {
        Random random = new Random();
        int d = m_angles.length;
        double[] shift = new double[d];
        double shift_norm = 0;
        for (int i = 0; i < d; i++)
        {
            shift[i] = random.nextDouble();
            shift_norm += shift[i] * shift[i];
        }
        shift_norm = Math.sqrt(shift_norm / this.distanceSquare(cluster.get(random.nextInt(cluster.size()))));
        double[] shift_minus = new double[d];
        for (int i = 0; i < d; i++)
        {
            shift[i] /= shift_norm;
            shift_minus[i] = - shift[i];
        }

        ClusterableTorus child1 = (ClusterableTorus)this.clone();
        ClusterableTorus child2 = (ClusterableTorus)this.clone();
        List<Clusterable> out = new ArrayList<>();
        out.add(child1.move(shift));
        out.add(child2.move(shift_minus));
        return out;
    }

    public double[] projection1(List<Clusterable> center, List<Clusterable> cluster)
    {
        if (center == null || center.size() != 1 || cluster == null || cluster.isEmpty())
        {
            return null;
        }
        final int n = cluster.size();
        final ClusterableTorus centroid = (ClusterableTorus) center.get(0);
        double[] out = new double[n];
        double out_mean = 0.0,
               out_sigma = 0.0;
        for (int i = 0; i < n; i++)
        {
            out[i] = Math.sqrt(centroid.distanceSquare(cluster.get(i)));
            out_mean += out[i];
            out_sigma += out[i] * out[i];
        }
        out_mean /= (double)n;
        out_sigma = Math.sqrt((out_sigma - out_mean * out_mean * n) / (double)(n - 1));

        for (int i = 0; i < n; i++)
        {
            out[i] = (out[i] - out_mean) / (out_sigma * (1 + EPS));
        }
        return out;
    }

    public double[] projection(List<Clusterable> direction, List<Clusterable> cluster)
    {
        if (direction == null || direction.size() != 2 || cluster == null || cluster.isEmpty())
        {
            return null;
        }
        long[][] x = new long[][]{((ClusterableTorus) direction.get(0)).m_angles,
                                  ((ClusterableTorus) direction.get(1)).m_angles};
        long[] center = new long[x[0].length];
        for (int i = 0; i < center.length; i++)
        {
            center[i] = Calc.circularMean(new long[]{x[0][i], x[1][i]}, PI);
        }
        long[] line = new long[center.length];
        double line_length_square = 0.0;
        for (int i = 0; i < line.length; i++)
        {
            line[i] = x[1][i] - x[0][i];
            line[i] = (line[i] >  PI ? line[i] - 2 * PI :
                      (line[i] < -PI ? line[i] + 2 * PI : line[i]));
            line_length_square += line[i] * line[i];
        }

        final int n = cluster.size();
        double[] out = new double[n];
        double out_mean = 0.0,
               out_sigma = 0.0;
        // This is the projection loop that might take long.
        // It scales as number_of_points * dimension.
        for (int i = 0; i < n; i++)
        {
            long[] that = ((ClusterableTorus) cluster.get(i)).m_angles;
            double scalar_product = 0.0;
            for (int j = 0; j < that.length; j++)
            {
                long tmp = that[j] - center[j];
                tmp = (tmp >  PI ? tmp - 2 * PI : (tmp < -PI ? tmp + 2 * PI : tmp));
                scalar_product += tmp * line[j];
            }
            out[i] = scalar_product / line_length_square;
            out_mean += out[i];
            out_sigma += out[i] * out[i];
        }
        out_mean /= (double)n;
        out_sigma = Math.sqrt((out_sigma - out_mean * out_mean * n) / (double)(n - 1));

        for (int i = 0; i < n; i++)
        {
            out[i] = (out[i] - out_mean) / (out_sigma * (1 + EPS));
        }
        return out;
    }

    private Clusterable move(double[] shift)
    {
        for (int i = 0; i < m_angles.length; i++)
        {
            m_angles[i] = Math.round((m_angles[i] + shift[i] + 2 * PI) % (2 * PI));
        }
        return this;
    }

    private Clusterable setRandom()
    {
        Random random = new Random();
        int d = m_angles.length;
        for (int i = 0; i < d; i++)
        {
            m_angles[i] = Math.round(random.nextDouble() * 360.0 * Const.MF);
        }
        return this;
    }

    @Override
    public Clusterable clone()
    {
        return new ClusterableTorus(m_angles, m_label);
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + Arrays.hashCode(m_angles);
        return result;
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        if (obj == null || getClass() != obj.getClass())
        {
            return false;
        }
        ClusterableTorus other = (ClusterableTorus) obj;
        if (m_label != other.m_label || !Arrays.equals(m_angles, other.m_angles))
        {
            return false;
        }
        return true;
    }

}
