package types;
/*
 * This file is part of CANE - A Clustering Algorithm for Non-Euclidean spaces.
 *
 * Copyright (C) 2014 Benjamin Eltzner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Scanner;

import main.CANE;
import clustering.Clusterable;
import clustering.ClusterableFromSimple;

public abstract class ImportData
{

    private static final int TORUS = 0, RESH = 1, SPHERE = 2, SERIES = 3;
    
    public static List<Clusterable> importCSV(int type, File csv_file)
    {
        if (csv_file == null || !csv_file.isFile() || !csv_file.canRead())
        {
            return null;
        }
        Map<String,List<String>> data = new HashMap<>();
        data.put("data_points", new ArrayList<String>());
        try
        {
            Scanner scanner = new Scanner(csv_file);
            while (scanner.hasNext())
            {
                if (scanner.hasNext("#.*"))
                {
                    if (CANE.DEBUG())
                    {
                        System.out.println("Skipping line starting with #.");
                    }
                    scanner.nextLine();
                }
                else if (scanner.hasNext("%[a-zA-Z_]*"))
                {
                    if (CANE.DEBUG())
                    {
                        System.out.println("Reading specifier line starting with %.");
                    }
                    String id = scanner.nextLine().substring(1);
                    data.put(id, new ArrayList<String>());
                    while (scanner.hasNext("%[^a-zA-Z]*"))
                    {
                        if (CANE.DEBUG())
                        {
                            System.out.println("Reading data line starting with %.");
                        }
                        data.get(id).add(scanner.nextLine().substring(1));
                    }
                }
                else 
                {
                    if (CANE.DEBUG())
                    {
                        System.out.println("Reading line.");
                    }
                    data.get("data_points").add(scanner.nextLine());
                }
            }
            double[][] matrix = null;
            double[] center = null;
            double radius = -1;
            int counter = 0;
            List<String> data_strings = null;
            String dec = Character.toString((new DecimalFormatSymbols()).getDecimalSeparator());
            for (String identifier : data.keySet())
            {
                switch (identifier)
                {
                case "matrix":
                    List<double[]> list = new ArrayList<>();
                    for (String line : data.get(identifier))
                    {
                        list.add(stringToArray(line, dec, 1));
                    }
                    matrix = new double[list.size()][];
                    for (int i = 0; i < matrix.length; i++)
                    {
                        matrix[i] = list.get(i);
                    }
                    break;
                case "center":
                    list = new ArrayList<>();
                    for (String line : data.get(identifier))
                    {
                        list.add(stringToArray(line, dec, 1));
                    }
                    if (list.size() == 1)
                    {
                        center = list.get(0);
                    }
                    break;
                case "radius":
                    list = new ArrayList<>();
                    for (String line : data.get(identifier))
                    {
                        list.add(stringToArray(line, dec, 1));
                    }
                    if (list.size() == 1 && list.get(0).length == 1)
                    {
                        radius = list.get(0)[0];
                    }
                    break;
                case "data_points":
                    data_strings = data.get(identifier);
                    break;
                default:
                    System.err.println("Unknown data identifier: " + identifier);
                }
            }
            Subsphere.set(matrix, center, radius);
            List<Clusterable> data_points = new ArrayList<>();
            switch (type)
            {
            case TORUS:
                for (String line : data_strings)
                {
                    double factor = Const.MF; // / Const.RAD;
                    double[] tmp = stringToArray(line, dec, factor);
                    if (tmp != null)
                    {
                        data_points.add(new ClusterableTorus(Calc.arrayRound(tmp), counter));
                        counter++;
                    }

                }
                break;
            case RESH:
                for (String line : data_strings)
                {
                    double factor = Const.MF;
                    double[] tmp = stringToArray(line, dec, factor);
                    if (tmp != null)
                    {
                        data_points.add(new ClusterableFromSimple(
                                        new ClusterableRESH(Calc.arrayRound(tmp), counter)));
                        counter++;
                    }
                                    
                }
                break;
            case SPHERE:
                for (String line : data_strings)
                {
                    double factor = Const.MF;
                    double[] tmp = stringToArray(line, dec, factor);
                    if (tmp != null)
                    {
                        data_points.add(new ClusterableFromSimple(
                                        new ClusterableSphere(Calc.arrayRound(tmp), counter)));
                        counter++;
                    }
                                    
                }
                break;
            case SERIES:
                for (String line : data_strings)
                {
                    double factor = Const.MF;
                    double[] tmp = stringToArray(line, dec, factor);
                    if (tmp != null)
                    {
                        data_points.add(new ClusterableTimeSeries(Calc.arrayRound(tmp), counter));
                        counter++;
                    }

                }
                break;
            default:
                System.err.println("Unknown data type: " + type);
            }
            scanner.close();
            return data_points;
        }
        catch(FileNotFoundException e)
        {
            System.out.println("File not found." + e.getMessage());
        }

        return null;
    }

    /**
     * Save cluster data to a csv file and do a cleanup.
     * @param file          The file to which to save.
     * @param cluster_data  The cluster segmentation to save.
     */
    public static void exportCSV (File file, Map<Clusterable, List<Clusterable>> cluster_data)
    {
        String export = "";
        for (Clusterable center : cluster_data.keySet())
        {
            export += "%" + center.exportCenter() + "\n";
            List<Clusterable> cluster = cluster_data.get(center);
            for (Clusterable item: cluster)
            {
                export += item.label() + ",";
            }
            export = export.substring(0, export.length()-1) + "\n";
        }
        writeFile(file, export);
        cleanup();
    }

    /**
     * Save cluster data to a csv file and do a cleanup.
     * @param file          The file to which to save.
     * @param cluster_data  The cluster segmentation to save.
     */
    public static void exportCSV (File file, List<List<Clusterable>> cluster_data)
    {
        String export = "";
        for (List<Clusterable> cluster : cluster_data)
        {
            for (Clusterable item: cluster)
            {
                export += item.label() + ",";
            }
            export = export.substring(0, export.length()-1) + "\n";
        }
        writeFile(file, export);
        cleanup();
    }

    public static Map<String, Integer> modes()
    {
        HashMap<String, Integer> map = new HashMap<String, Integer>();
        map.put("Original angular data", TORUS);
        map.put("Points on RESH space", RESH);
        map.put("Points on sphere", SPHERE);
        map.put("Time Series", SERIES);
        return map;
    }

    public static String type(int type)
    {
        switch (type)
        {
        case TORUS:
            return "torus";
        case RESH:
            return "RESH";
        case SPHERE:
            return "sphere";
        case SERIES:
            return "time series";
        }
        return "other";
    }

    private static void cleanup()
    {
        Subsphere.clear();
    }

    private static boolean writeFile(File file, String content)
    {
        try
        {
            if (file == null)
            {
                // "[A-Za-z0-9]{3,4}$" should allow for 3 or 4 character file
                // name extensions.
                //
                return false;
            }

            BufferedWriter bwr = new BufferedWriter(new FileWriter(file, false));
            bwr.write(content);
            bwr.close();

            if (CANE.DEBUG())
            {
                System.out.println("Exported data to '" + file + "'.");
            }
        }
        catch (IOException e)
        {
            System.err.println("Exporting to file failed: " + e);
            return false;
        }
        return true;
    }

    private static double[] stringToArray(String line, String dec, double factor)
    {
        line = line.replaceAll(",", "_").replaceAll("\\.", dec);
        Scanner scanner = new Scanner(line);
        scanner.useDelimiter("_");
        List<Double> list = new ArrayList<>();
        while (scanner.hasNextDouble())
        {
            list.add(scanner.nextDouble() * factor);
        }
        scanner.close();
        double[] out = new double[list.size()];
        for (int i = 0; i < out.length; i++)
        {
            out[i] = list.get(i);
        }
        if (out.length == 0)
        {
            return null;
        }
        return out;
    }
}
