package types;
/*
 * This file is part of CANE - A Clustering Algorithm for Non-Euclidean spaces.
 *
 * Copyright (C) 2014 Benjamin Eltzner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


/**
 * Simplistic singleton for the conversion between the two representations
 * of points on the RESH space.
 */
class Subsphere
{

    public static final Subsphere the = new Subsphere();

    private double[][] m_matrix = null;
    private double[] m_center = null;
    private double m_radius = -1;
    private int m_n_angles,
                m_n_coordinates;
    

    private Subsphere(){}
    
    public static void set (double[][] matrix, double[] center, double radius)
    {
        if (the.m_matrix != null && the.m_center != null && the.m_radius >= 0)
        {
            System.err.println("Sphere has already been initialized.");
            return;
        }
        the.m_matrix = matrix;
        the.m_center = center;
        the.m_radius = radius;
        if (matrix != null)
        {
            the.m_n_angles = matrix.length - 1;
        }
        if (center != null)
        {
            the.m_n_coordinates = center.length;
        }
    }

    public static void clear()
    {
        the.m_matrix = null;
        the.m_center = null;
        the.m_radius = -1;
    }

    public static double[] vectorToAngles(double[] vector)
    {
       return v2a(project(vector));
    }

    public static double[] anglesToVector(double[] angles)
    {
        return unproject(a2v(angles));
    }

    public static double[] project(double[] vector)
    {
        if (the.m_matrix == null && the.m_center == null && the.m_radius < 0)
        {
            System.err.println("Sphere has not been initialized.");
            return null;
        }
        if (vector == null)
        {
            return null;
        }

        double[] projection = new double[the.m_n_angles+1];
        for (int i = 0; i <= the.m_n_angles; i++)
        {
            for (int j = 0; j < the.m_n_coordinates; j++)
            {
                projection[i] += the.m_matrix[i][j] * vector[j] / the.m_radius;
            }
        }
        return projection;
    }

    public static double[] unproject(double[] vector)
    {
        if (the.m_matrix == null && the.m_center == null && the.m_radius < 0)
        {
            System.err.println("Sphere has not been initialized.");
            return null;
        }
        if (vector == null)
        {
            return null;
        }

        double[] out = new double[the.m_n_coordinates];
        
        for (int j = 0; j < the.m_n_coordinates; j++)
        {
            out[j] = Const.MF * the.m_center[j];
            for (int i = 0; i <= the.m_n_angles; i++)
            {
                out[j] += the.m_matrix[i][j] * vector[i] * the.m_radius;
            }
        }
        return out;
    }
    
    public static double[] v2a (double[] vector)
    {
        return v2a(vector, the.m_n_angles);
    }

    public static double[] a2v (double[] angles)
    {
        if (angles == null)
        {
            return null;
        }

        angles = Calc.projectRESH(angles);
        double[] vector = new double[the.m_n_angles+1];
        for (int i = 0; i < the.m_n_angles; i++)
        {
            vector[i] = Const.MF * Math.cos(angles[i] * Const.RAD / Const.MF);
            for (int j = 0; j < i; j++)
            {
                vector[i] *= Math.sin(angles[j] * Const.RAD / Const.MF);
            }
        }
        vector[the.m_n_angles] = Const.MF;
        for (int j = 0; j < the.m_n_angles; j++)
        {
            vector[the.m_n_angles] *= Math.sin(angles[j] * Const.RAD / Const.MF);
        }
        return vector;
    }

    public static double[] allAngles(double[] angles)
    {
        return v2a(anglesToVector(angles), the.m_n_coordinates-1);
    }

    public static double radius()
    {
        return the.m_radius;
    }

    private static double[] v2a(double[] vector, int n)
    {
        if (vector == null)
        {
            return null;
        }

        double[] angles = new double[n];
        for (int i = 0; i < n; i++)
        {
            for (int j = 0; j < i; j++)
            {
                vector[i] /= Math.sin(angles[j]);
            }
            angles[i] = Math.acos(vector[i] / Const.MF);
        }
        for (int i = 0; i < n; i++)
        {
            angles[i] = angles[i] * Const.MF / Const.RAD;
        }
        if (Math.asin(vector[n] / Const.MF) < 0)
        {
            angles[n-1] = 360 * Const.MF - angles[n-1];
        }
        return angles;
    }
}
