# -*- coding: utf-8 -*-
"""
Copyright (C) 2014 Benjamin Eltzner

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
or see <http://www.gnu.org/licenses/>.
"""

import matplotlib
try: matplotlib.use('Agg')
except: print('Rerunning')
import numpy as np
import numpy.linalg as la
import math, sys
from RESH_io import find_files, import_csv, import_lists, export_csv
from RESH_plot import (scatter_plots, var_plot, inv_var_plot, residual_plots,
                       sphere_views, make_circle, colored_scatter_plots,
                       one_scatter_plot, abs_var_plot, residues_plot, rainbow,
                       plot_thread)
from RESH_geometry import RESHify_1D, unRESHify_1D, torus_distances, euclideanize
from RESH_PNS  import pns_loop, unfold_points, as_matrix, two_sample_test
from Multiscale_modes import get_quantile, get_modes

"""
AUTHOR'S DISCLAIMER:
This software was written in view of a specific data set on a specific system
and is provided for the sake of verifiability of its results. It is not
recommended to reuse this software in any other scenario.

The functions in this program do not check for well-formed input which may lead
to unexpected or even harmful behavior. Output functions do not check for
pre-existing data. Running the program CAN RESULT IN LOSS OF DATA by unchecked
overwriting and CAN POTENTIALLY DAMAGE YOUR SYSTEM.

This disclaimer need not be reproduced in copies or derived works of this
software. However, the author would be grateful if it was kept intact at least
in unmodified copies.
"""

################################################################################
################################   Constants   #################################
################################################################################

DEG = math.degrees(1)

################################################################################
#############################   Control function   #############################
################################################################################

""" Full data processing """
def main_all ():
    points = import_csv(find_files('RNA_data.csv')[0])['PDB-Data'][:,:7]
    print(points.shape)
    val, vec = la.eig(np.cov(euclideanize(points.copy()).T))
    val = np.hstack((np.zeros(1), np.cumsum(np.sort(val)[::-1])))
    max_var = val[-1]
    val = max_var - val
    scree_data = []
    for invert in [False, True]:
        name = ('SO' if invert else 'SI')
#        scatter_plots(points, 'full_data_' + name + '_codim0')
        sphere_points, means, half = RESHify_1D(points, invert)
        spheres, projected_points, distances = pns_loop(sphere_points, True, True, mode='scale')
        for s in spheres[:-1]: print(s.radius)
        residual_plots(distances, 1, 3, 'blue', 'full_data_resh_residuals_' + name)
#        make_bar_chart(make_binning(projected_points[-2], True),
#                       filename=('full_data_' + name + '_histogram'))
        center_point = projected_points[-1]
        center_point = unRESHify_1D(unfold_points(as_matrix(center_point),
                                                      spheres[:-1]), means, half)
        unfolded_angles = [unRESHify_1D(unfold_points(pts, spheres[:i+1]), means, half)
                           for i, pts in enumerate(projected_points[:-1])]
#        old_ang = points
        var_data = []
        for j, ang in enumerate(unfolded_angles):
            var_data.append(np.mean(torus_distances(ang, points)**2))
#            old_ang = ang
#            scatter_plots(ang, 'full_data_' + name + '_codim%d' % (j+1))
        full_var = np.mean(torus_distances(center_point, points)**2)
        tmp = max([full_var] + var_data)
        if tmp > max_var:
            max_var = tmp
        scree_data.append(np.array([[0,full_var]] + [[i+1,v] for i,v in enumerate(reversed(var_data))]))
    scree_data = [np.array([[i, v] for i,v in enumerate(val[:-1])])] + scree_data
    for i in range(len(scree_data)):
        scree_data[i][:,1] *= 100 / max_var
    print(max_var)
    inv_var_plot(scree_data, ('Tangent', 'SI', 'SO'), 'full_data_resh_scree')
    return max_var

""" Clustering """
def new_multi_slink (scale):
    points = import_csv(find_files('RNA_data.csv')[0])['PDB-Data']
    clusters = import_lists(find_files('RNA_data*multiSLINK_result.csv')[0])
    noise = [np.array(sorted(clusters[-1]))]
    clusters = [np.array(sorted(x)) for x in clusters[:-1]]
    scree_data = []
    scree_data_euclid = []
    std_data = []
    scree_labels = []
    clusters = __slink_pns (clusters, points, scree_data, scree_data_euclid, std_data, scree_labels, 'filtered', scale)
    noise = __slink_pns (noise, points, scree_data, scree_data_euclid, std_data, scree_labels, 'noise', scale)
    tmp = [[rainbow(i,len(clusters)), points[c]] for i,c in enumerate(clusters)]
    colored_scatter_plots(tmp, 0, 7, 'full_data_slink_clusters_7d')
    colored_scatter_plots(tmp, 7, 9, 'full_data_slink_clusters_2d')
    export_csv({scree_labels[i]: x for i,x in enumerate(std_data)},
               'slink_std_data.csv', mode='Float')
    var_plot(scree_data, 90, scree_labels, False, 'full_data_slink_scree')
    var_plot(scree_data_euclid, 90, scree_labels, False, 'full_data_slink_scree_euclidean')
    abs_var_plot(std_data, 10, scree_labels, 'standard deviation [${}^\circ$]',
                 'full_data_slink_std')
    for i,eu in enumerate(scree_data_euclid):
        name = 'scree_' + scree_labels[i].replace('_',' ')
        var_plot([eu, scree_data[i]], 80, ['Euclid', 'PNS'], False, name)

def __slink_pns (new_clusters, points, scree_data, scree_data_euclid, std_data,
                 scree_labels, type_name, scale):
    new_clusters = list(reversed(sorted(new_clusters, key=len)))
    split = True
    count = 0
    final_clusters = []
    while split:
        split = False
        count += 1
        clusters = list(reversed(sorted(new_clusters, key=len)))
        if type_name == 'filtered':
            export_csv({('cluster%02d'%i): c for i,c in enumerate(clusters)},
                        'slink_clusters_1d.csv', mode='Int')
        elif type_name == 'noise':
            export_csv({('cluster%02d'%i): c for i,c in enumerate(clusters)},
                        'slink_noise_1d.csv', mode='Int')
        new_clusters = []
        for i,c in enumerate(clusters):
            if any(np.array_equal(c, f) for f in final_clusters):
                new_clusters.append(c)
                continue
            name = 'slink_' + type_name + '_run%d_cluster%02d_%d_points_' % (count, i, len(c))
            p = points[c]
            print(p.shape)
            quantile = get_quantile(len(p))
            for [invert, mode] in [[False, 'gap'], [True, 'gap'],
                                   [False, 'mean'], [True, 'mean']]:
                suffix = ('SO_' if invert else 'SI_') + mode + '_'
                sphere_points, means, half = RESHify_1D(p[:,:7], invert, mode)
                spheres, projected_points, distances = pns_loop(sphere_points, degenerate=half, verbose=False, mode='scale')
                for s in spheres[:-1]: print(s.radius)
                sys.stdout.flush()
                mode_list, mins = get_modes(distances[-1], 360., quantile)
                print('SO' if invert else 'SI', mode, mins)
                sys.stdout.flush()
                if len(mins) > 1:
                    __slink_plotting(name, suffix, p, distances, projected_points,
                                     spheres, means, half, new_clusters, mode_list,
                                     c, count, scree_data, scree_data_euclid,
                                     std_data, scree_labels, False, scale)
                    split = True
                    break
            if len(mins) <= 1:
                __slink_plotting(name, suffix, p, distances, projected_points,
                                 spheres, means, half, new_clusters, mode_list,
                                 c, count, scree_data, scree_data_euclid,
                                 std_data, scree_labels, True, scale)
                final_clusters.append(c)
    print('Clustering done.', len(clusters), 'clusters found.')
    sys.stdout.flush()
    return clusters

def __slink_plotting (name, suffix, p, distances, projected_points, spheres,
                      means, half, new_clusters, mode_list, c, count, scree_data,
                      scree_data_euclid, std_data, scree_labels, stop, scale):
    if stop:
        name = name.replace('_run', '_stop_run')
    def worker ():
        scatter_plots(p[:,:7], name + '7d')
        scatter_plots(p[:,7:9], name + '2d')
        residual_plots(distances, 1, 4, 'blue', name + suffix + 'residuals')
        angles_1d = unRESHify_1D(unfold_points(projected_points[-2], spheres[:-1]), means, half)
        scatter_plots(angles_1d, name + suffix + 'projection_1d')
    plot_thread(worker)
    new_clusters += [c[x] for x in mode_list]
    if (count == 1):
        center_point = unRESHify_1D(unfold_points(as_matrix(projected_points[-1]), spheres[:-1]), means, half)
        unfolded_angles = [unRESHify_1D(unfold_points(pts, spheres[:i+1]), means, half) for i, pts in enumerate(projected_points[:-1])]
        var_data = []
        for ang in unfolded_angles:
            var_data.append(np.mean(torus_distances(ang, p[:,:7])**2))
        var_data.append(np.mean(torus_distances(center_point, p[:,:7])**2))
        scree_data.append(np.array([[i, 100 * (1 - v / scale)] for i,v in enumerate(reversed(var_data))]))
        std_data.append(np.array([[i, math.sqrt(v)] for i,v in enumerate(reversed(var_data))]))
        scree_labels.append(('noise ' if 'noise' in name else '') + str(len(p)) +
                            ' ' + suffix[:-1].replace('_',' '))
        val, vec = la.eig(np.cov(euclideanize(p[:,:7].copy()).T))
        val = np.hstack((np.zeros(1), np.cumsum(np.sort(val)[::-1])))
        val = val[-1] - val
        scree_data_euclid.append(np.array([[i, 100 * (1 - v / scale)] for i,v in enumerate(val[:-1])]))

""" geoPCA """
def geoPCA_orig_data ():
    points = (import_csv(find_files('geoPCA_data.csv')[0]) + 360) % 360
    print(points.shape)
    def crop(v, bounds):
        return v[np.all([(v[:,a]>b)&(v[:,a]<c) for [a,b,c] in bounds], axis=0)]
    clusters = [[rainbow(0,3),crop(points,[[7, 15, 45],[8,225,255]])],
                [rainbow(1,3),crop(points,[[7,299,329],[8,216,246]])],
                [rainbow(2,3),crop(points,[[7,147,187],[8,330,360]])]]
    print([c[1].shape for c in clusters])
    one_scatter_plot(clusters, 0, 5, [r'$\alpha$ [${}^\circ$]',r'$\zeta$ [${}^\circ$]'], 'geoPCA_orig_labels_7d_az')
    one_scatter_plot(clusters, 7, 8, [r'$\eta$ [${}^\circ$]',r'$\theta$ [${}^\circ$]'], 'geoPCA_orig_labels_2d')
    lengths = [0] + [c[1].shape[0] for c in clusters]
    colors = sum(([rainbow(i,3)]*n for i,n in enumerate(lengths[1:])), [])
    points = np.vstack([c[1] for c in clusters])
    for invert in [False, True]:
        new_clusters = [np.array(range(len(points)))]
        split = True
        count = 0
        while split:
            split = False
            count += 1
            clusters = list(reversed(sorted(new_clusters, key=len)))
            export_csv({'cluster'+str(i): c for i,c in enumerate(clusters)},
                       'geoPCA_orig_clusters_1d.csv', mode='Int')
            new_clusters = []
            for i,c in enumerate(clusters):
                name = 'geoPCA_run%d_cluster_%d_%d_points_' % (count, i, len(c))
                name += ('SO_' if invert else 'SI_')
                p = points[c]
                print(p.shape)
                quantile = get_quantile(len(p))
                sphere_points, means, half = RESHify_1D(p[:,:7], invert, mode='mean')
                spheres, projected_points, distances = pns_loop(sphere_points, degenerate=half, verbose=True, mode='scale')
                for s in spheres[:-1]: print(s.radius)
                sys.stdout.flush()
                if len(p) == len(points):
                    if not invert:
                        projected_points[-3][:,-1] *= -1
                        spheres[-2].normal[-1] *= -1
                    circle = make_circle(spheres[-2].normal, spheres[-2].height)
                    sphere_views(projected_points[-3], spheres[-2].normal, colors, circle, name[:-1])
                if count == 1:
                    residual_plots(distances, 1, 3, colors, 'geoPCA_orig_residuals')
#                scatter_plots(p[:,:7], name + '7d')
#                scatter_plots(p[:,7:9], name + '2d')
#                residual_plots(distances, 1, 3, 'blue', name + 'resh_residuals')
                mode_list, mins = get_modes(distances[-1], 360., quantile)
                print(mins)
                sys.stdout.flush()
                if len(mins) <= 1:
                    new_clusters.append(c)
                    continue
                new_clusters += [c[x] for x in mode_list]
                split = True
        print('Clustering done.', len(clusters), 'clusters found.')
        sys.stdout.flush()
        tmp = [[rainbow(i,(3 if len(clusters)==3 else 1.5)),
                points[c]] for i,c in enumerate(clusters)]
        kind = ('SO' if invert else 'SI')
        one_scatter_plot(tmp, 0, 5, [r'$\alpha$ [${}^\circ$]',r'$\zeta$ [${}^\circ$]'],
                         'geoPCA_orig_found_clusters_7d_az_' + kind)
        one_scatter_plot(tmp, 7, 8, [r'$\eta$ [${}^\circ$]',r'$\theta$ [${}^\circ$]'],
                         'geoPCA_orig_found_clusters_2d_' + kind)

def scree ():
    stds = import_csv(find_files('slink_std_data.csv')[0])
    scree_data = []
    scree_labels = []
    scale = 13742.1871427
    for key in stds:
        scree_labels.append(key)
        tmp = stds[key]
        tmp[:,1] = 100 * (1 - tmp[:,1]**2 / scale)
        scree_data.append(tmp)
    var_plot(scree_data, 90, scree_labels, False, 'full_data_slink_scree_alt')

def plots_2d_vs_7d ():
    points = import_csv(find_files('RNA_data.csv')[0])['PDB-Data']
    tmp = import_lists(find_files('slink_clusters_1d.csv')[0])
    clusters = []
    for key in tmp:
        c = tmp[key]
        if not len(c) in [4921, 477, 137]:
            continue
        clusters.append(points[c])
    clusters = list(reversed(sorted(clusters, key=len)))
    tmp = [[rainbow(i,len(clusters)), p] for i,p in enumerate(clusters)]
#    [a, b] = clusters
#    distances = np.zeros((len(a), len(b)))
#    for i in range(7):
#        pa = a[:,i]
#        pb = b[:,i]
#        d = np.abs(pb - pa[:,np.newaxis])
#        d[d>180] -= 360
#        distances += d**2
#    distances = np.min(np.sqrt(distances), axis=0)
#    print(distances)
    one_scatter_plot(tmp, 0, 5, [r'$\alpha$ [${}^\circ$]',r'$\zeta$ [${}^\circ$]'], '3clusters_7d_az')
    one_scatter_plot(tmp, 7, 8, [r'$\eta$ [${}^\circ$]',r'$\theta$ [${}^\circ$]'], '3clusters_2d')
    colored_scatter_plots(tmp, 0, 7, '3clusters_7d')
    colors = sum(([col]*len(p) for [col, p] in tmp), [])
    print(len(colors))
    points = np.vstack([p for [_, p] in tmp])[:,:7]
    best_fit = None; proj2d = None; normal = None; circle = None; suffix = None
    for [invert, mode] in [[False, 'gap'], [True, 'gap']]:#,
                           #[False, 'mean'], [True, 'mean']]:
        print(('SO ' if invert else 'SI ') + mode)
        sphere_points, means, half = RESHify_1D(points, invert, mode)
        spheres, projected_points, distances = pns_loop(sphere_points, degenerate=half, verbose=False, mode='scale')
        for s in spheres[:-1]: print(s.radius)
        sys.stdout.flush()
        center_point = unRESHify_1D(unfold_points(as_matrix(projected_points[-1]), spheres[:-1]), means, half)
        unfolded_angles = [unRESHify_1D(unfold_points(pts, spheres[:i+1]), means, half) for i, pts in enumerate(projected_points[:-1])]
        var_data = []
        for ang in unfolded_angles:
            var_data.append(np.mean(torus_distances(ang, points)**2))
        var_data.append(np.mean(torus_distances(center_point, points)**2))
        var_data = np.array(var_data)[::-1]
        print(var_data)
        sys.stdout.flush()
        if (best_fit == None) or (var_data[1] < best_fit[1]):
            print('better fit!')
            best_fit = var_data
            proj2d = projected_points[-3]
            normal = spheres[-2].normal
            circle = make_circle(spheres[-2].normal, spheres[-2].height)
            suffix = ('SO_' if invert else 'SI_') + mode
    sphere_views(proj2d, normal, colors, circle, '3clusters_' + suffix)

def sec_struct_for_3 ():
    tmp = import_lists(find_files('slink_clusters_1d.csv')[0])
    tmp = list(reversed(sorted([tmp[k] for k in tmp], key=len)))
    names = import_lists(find_files('RNA_data_with_meta.csv')[0], as_type='')
    points = np.array([n[:-1] for n in names]).astype(np.float)
    names = np.array([n[-1] for n in names])
    pairs = import_lists(find_files('RNA_data_pairs.csv')[0], as_type='')
    pair_dict = {}
    for [me, t, they] in pairs:
        if not me in pair_dict:
            pair_dict[me] = []
        pair_dict[me].append([t, they])
    clusters = []
    name_list = []
    pair_types = []
    for c in tmp:
        if not len(c) in [4921, 477, 137]:
            continue
        print(len(c), np.mean(points[c,:7], axis=0))
        sugar_count = len(points[points[c,-1] < 180.])
        multiplicities = [0,0,0,0,0,0]
        these_names = names[c]
        base_dict = {'A' : 0, 'C' : 0, 'G' : 0, 'T' : 0, 'U' : 0}
        residue_dict = {}
        pair_type_dict = {}
        for name in these_names:
            base_dict[name[-1]] += 1
            if not name[:4] in residue_dict:
                residue_dict[name[:4]] = {}
            if not name[:4] in name_list:
                name_list.append(name[:4])
            if not name[4] in residue_dict[name[:4]]:
                residue_dict[name[:4]][name[4]] = []
            residue_dict[name[:4]][name[4]].append([int(name[5:9]), name[9]])
            this_pair_types = (['none'] if not name in pair_dict
                               else [p[0]+name[-1]+p[1][-1] for p in pair_dict[name]])
            if name in pair_dict:
                multiplicities[len(pair_dict[name])] += 1
            else:
                multiplicities[0] += 1
            for t in this_pair_types:
                if not t in pair_type_dict:
                    pair_type_dict[t] = []
                if not t in pair_types:
                    pair_types.append(t)
                pair_type_dict[t].append(name)
        clusters.append([len(c), sugar_count, base_dict, residue_dict,
                         pair_type_dict, multiplicities])
    test = 'Sugars '
    for [n, s, _, _, _, _] in clusters:
        test += '%.2f \\%% ' % (100 * s / n)
    print(test)
    for base in ['A', 'C', 'G', 'U']:
        test = base + ' '
        for [n, _, b, _, _, _] in clusters:
            test += '%.2f \\%% ' % (100 * b[base] / n)
        print(test)
    for name in sorted(name_list):
        test = name + ' '
        for [_, _, _, c, _, _] in clusters:
            if not name in c:
                test += '0 '
                continue
            d = c[name]
            test += str(sum([len(d[key]) for key in d])) + ' '
        print(test)
    for pair_type in sorted(pair_types):
        test = pair_type + ' '
        common = False
        for i, [n, _, _, _, c, _] in enumerate(clusters):
            if not pair_type in c:
                test += '0.00 \% '
                continue
            d = c[pair_type]
            test += '%.2f \\%% ' % (100 * len(d) / n)
            if len(d) > n / 100:
                common = True
        if common:
            print(test)
    for k in range(6):
        test = 'Multiplicity %d ' % k
        for [n, _, _, _, _, mult] in clusters:
            test += '%.2f \\%% ' % (100 * mult[k] / n)
        print(test)

def paper_plots ():
    points = import_csv(find_files('RNA_data.csv')[0])['PDB-Data'][:,:7]
    print(points.shape)
#    val, vec = la.eig(np.cov(euclideanize(points.copy(), mode='mean').T))
#    val = np.hstack((np.zeros(1), np.cumsum(np.sort(val)[::-1])))
#    val = np.max(val) - val
#    max_var = val[0]
    max_var = val = 0
    max_var = __full_data_plot (max_var, points, val)
    #__slink_plots(max_var, points)

def __full_data_plot (max_var, points, val):
    scree_data = []
    for invert in [False, True]:
        sphere_points, means, half = RESHify_1D(points, invert)
        spheres, projected_points, distances = pns_loop(sphere_points, True, True, mode='scale')
        for s in spheres[:-1]: print(s.radius)
        center_point = projected_points[-1]
        center_point = unRESHify_1D(unfold_points(as_matrix(center_point),
                                                      spheres[:-1]), means, half)
        unfolded_angles = [unRESHify_1D(unfold_points(pts, spheres[:i+1]), means, half)
                           for i, pts in enumerate(projected_points[:-1])]
        var_data = []
        for j, ang in enumerate(unfolded_angles):
            var_data.append(np.mean(torus_distances(ang, points)**2))
        full_var = np.mean(torus_distances(center_point, points)**2)
        tmp = max([full_var] + var_data)
        if tmp > max_var:
            max_var = tmp
        scree_data.append(np.array([[0,full_var]] + [[i+1,v] for i,v in enumerate(reversed(var_data))]))
    #scree_data = [np.array([[i, v] for i,v in enumerate(val[:-1])])] + scree_data
    for i in range(len(scree_data)):
        scree_data[i][:,1] *= 100 / max_var
    print(max_var)
    scree_data = [np.vstack((x,np.array([x[-1,0]+1,0]))) for x in scree_data]
    inv_var_plot(scree_data, ('SI', 'SO'), 'full_data_resh_scree')
    return max_var

def __slink_plots (max_var, points):
    clusters = import_lists(find_files('RNA_data*multiSLINK_result.csv')[0])
    clusters = [np.array(sorted(x)) for x in clusters[:-1]]
    scree_data = []
    scree_data_euclid = []
    scree_labels = []
    euclid_labels = []
    clusters = list(reversed(sorted(clusters, key=len)))
    for i,c in enumerate(clusters):
        if len(c) < 50:
            continue
        p = points[c]
        val, vec = la.eig(np.cov(euclideanize(p.copy(), mode='mean').T))
        val = np.hstack((np.zeros(1), np.cumsum(np.sort(val)[::-1])))
        val = np.max(val) - val
        if val[1] < 0.1 * max_var:
            continue
        name = 'cluster%02d_%d_points_' % (i, len(c))
        print(p.shape)
        best_fit = None; proj2d = None; normal = None; circle = None; suffix = None
        for [invert, mode] in [[False, 'gap'], [True, 'gap'],
                               [False, 'mean'], [True, 'mean']]:
            print(('SO ' if invert else 'SI ') + mode)
            sphere_points, means, half = RESHify_1D(p, invert, mode)
            spheres, projected_points, distances = pns_loop(sphere_points, degenerate=half, verbose=False, mode='scale')
            for s in spheres[:-1]: print(s.radius)
            sys.stdout.flush()
            center_point = unRESHify_1D(unfold_points(as_matrix(projected_points[-1]), spheres[:-1]), means, half)
            unfolded_angles = [unRESHify_1D(unfold_points(pts, spheres[:i+1]), means, half) for i, pts in enumerate(projected_points[:-1])]
            var_data = []
            for ang in unfolded_angles:
                var_data.append(np.mean(torus_distances(ang, p)**2))
            var_data.append(np.mean(torus_distances(center_point, p)**2))
            var_data = np.array(var_data)[::-1]
            print(var_data)
            sys.stdout.flush()
            if (best_fit == None) or (var_data[1] < best_fit[1]):
                print('better fit!')
                best_fit = var_data
                proj2d = projected_points[-3]
                normal = spheres[-2].normal
                circle = make_circle(spheres[-2].normal, spheres[-2].height)
                suffix = ('SO_' if invert else 'SI_') + mode
        if val[1] > max(0.2 * max_var, best_fit[1] * 1.5):
            sphere_views(proj2d, normal, 'blue', circle, name + suffix)
            scree_data.append(np.array([[i, 100 * v / max_var] for i,v in enumerate(best_fit)]))
            scree_data_euclid.append(np.array([[i, 100 * v / max_var] for i,v in enumerate(val[:-1])]))
            scree_labels.append((name+suffix).replace('_',' '))
            euclid_labels.append((name[:-1]).replace('_',' '))
    inv_var_plot(scree_data, scree_labels, 'full_data_slink_scree')
    inv_var_plot(scree_data_euclid, euclid_labels, 'full_data_slink_scree_euclidean')

def t2_testing ():
    points = import_csv(find_files('RNA_data.csv')[0])['PDB-Data']
    tmp = import_lists(find_files('slink_clusters_1d.csv')[0])
    clusters = []
    for key in tmp:
        c = tmp[key]
        if not len(c) in [4921, 164]:
            continue
        clusters.append(points[c])
    clusters = list(reversed(sorted(clusters, key=len)))
    tmp = [[rainbow(i,len(clusters)), p] for i,p in enumerate(clusters)]
    one_scatter_plot(tmp, 0, 5, [r'$\alpha$ [${}^\circ$]',r'$\zeta$ [${}^\circ$]'])
    one_scatter_plot(tmp, 7, 8, [r'$\eta$ [${}^\circ$]',r'$\theta$ [${}^\circ$]'])
    data = []
    for c in clusters:
        sphere_points, means, half = RESHify_1D(c[:,:7], False, 'gap')
        data.append(sphere_points)
    print('starting test')
    print(two_sample_test(data[0], data[1]))

def check_files ():
    points = import_csv(find_files('RNA_data.csv')[0])['PDB-Data']
    names = import_lists(find_files('RNA_data_with_meta.csv')[0], as_type='')
    points2 = np.array([n[:-1] for n in names])
    coins = np.where(la.norm(points[:,:,np.newaxis] -
                             points2.T[np.newaxis,:,:], axis=1) < 1e-10)
    print(coins)

def plot_residues ():
    points = import_csv(find_files('RNA_data.csv')[0])['PDB-Data']
    tmp = import_lists(find_files('slink_clusters_1d.csv')[0])
    tmp = list(reversed(sorted([tmp[k] for k in tmp], key=len)))
    cluster_means = []
    for c in tmp:
        if not len(c) in [4921, 477, 137]:
            continue
        #means.append(np.mean(points[c,:6], axis=0))
        sphere_points, means, half = RESHify_1D(points[c,:6], False, 'gap')
        spheres, projected_points, distances = pns_loop(sphere_points, degenerate=half, verbose=False, mode='scale')
        for s in spheres[:-1]: print(s.radius)
        sys.stdout.flush()
        center_point = unRESHify_1D(unfold_points(as_matrix(projected_points[-1]), spheres[:-1]), means, half)
        print(len(c), np.rint(center_point.flatten()))
        cluster_means.append(center_point.flatten())
    colors = [rainbow(i,len(cluster_means)) for i in range(len(cluster_means))]
    residues_plot(cluster_means, colors, 3)#, 'residue_geometries')
#    helices = [np.array(55 * list(cluster_means[0]))] + [
#               np.array((60 // (n+1)) * (list(cluster_means[1]) + n * list(cluster_means[0])))
#               for n in reversed(range(5,11))]
#    colors = [rainbow(i,len(helices)) for i in range(len(helices))]
#    residues_plot(helices, colors, 9, 'helices')

def cluster_sizes ():
    clusters = import_lists(find_files('RNA_data*multiSLINK_result.csv')[0])
    for c in sorted(clusters, key=len):
        print(len(c))

def make_8 ():
    x = list(range(7)) + [9]
    points = import_csv(find_files('RNA_data.csv')[0])['PDB-Data'][:,x]
    export_csv({'PDB-Data': points},
               '../data/PDB/RNA_data_8angles.csv', mode='Float')

def check_chains ():
    points = import_csv(find_files('RNA_data.csv')[0])['PDB-Data']
    tmp = import_lists(find_files('slink_clusters_1d.csv')[0])
    tmp = list(reversed(sorted([tmp[k] for k in tmp], key=len)))
    chains = import_csv(find_files('RNA_chains_data.csv')[0])
    cluster_points = []
    words2 = []
    words3 = []
    for c in tmp:
        if not len(c) in [4921, 477, 137]:
            continue
        print(len(c))
        cluster_points.append(points[c,:6])
    for key in chains:
        if len(chains[key].shape) == 1:
            chain = chains[key][np.newaxis,:6].T
        else:
            chain = chains[key][:,:6].T
        signature = np.array(['*' for a in range(chain.shape[1])])
        for i, c in enumerate(cluster_points):
            dists = c[:,:,np.newaxis] - chain[np.newaxis,:,:]
            tmp = np.where(np.min(la.norm(dists, axis=1), axis=0) < 1e-6)[0]
            if i < 10:
                signature[tmp] = '%d' % i
            else:
                signature[tmp] = chr(i-10 + ord('A'))
        print(key, "".join(signature))
        words2 += [w for w in "".join(signature).split('*') if '2' in w[1:-1]]
        words3 += [w for w in "".join(signature).split('*') if '3' in w[1:-1]]
    words2 = [w for w in words2 if len(w) > 4]
    words3 = [w for w in words3 if len(w) > 4]
    print(words2, len(words2))
    print(words3, len(words3))

def cluster_overview ():
    points = import_csv(find_files('RNA_data.csv')[0])['PDB-Data']
    tmp = import_lists(find_files('slink_clusters_1d.csv')[0])
    tmp = list(reversed(sorted([tmp[k] for k in tmp], key=len)))
    for c in tmp:
        if not len(c) in [59,60,107,232]: continue
        p = points[c,:7]
        best_fit = None
        nested_mean = None
        best_id = None
        proj2d = None
        normal = None
        circle = None
        for [invert, mode] in [[False, 'gap'], [True, 'gap'],
                               [False, 'mean'], [True, 'mean']]:
            identifier = ('SO ' if invert else 'SI ') + mode
            sphere_points, means, half = RESHify_1D(p, invert, mode)
            spheres, projected_points, distances = pns_loop(sphere_points, degenerate=half, verbose=False, mode='scale')
            center_point = unRESHify_1D(unfold_points(as_matrix(projected_points[-1]), spheres[:-1]), means, half)
            unfolded_angles = [unRESHify_1D(unfold_points(pts, spheres[:i+1]), means, half) for i, pts in enumerate(projected_points[:-1])]
            var_data = []
            for ang in unfolded_angles:
                var_data.append(np.mean(torus_distances(ang, p)**2))
            var_data.append(np.mean(torus_distances(center_point, p)**2))
            var_data = np.array(var_data)[::-1]
            if (best_fit == None) or (var_data[1] < best_fit[1,1]):
                best_fit = np.array([[i, v] for i,v in enumerate(var_data)])
                nested_mean = center_point
                best_id = identifier
                proj2d = projected_points[-3]
                normal = spheres[-2].normal
                circle = make_circle(spheres[-2].normal, spheres[-2].height)
        sphere_views(proj2d, normal, 'blue', circle, 'slink_suspicious_cluster_'  + str(len(c)) + '_' + best_id)
        print(len(c), '&', best_id)
        print(nested_mean[0], best_fit[:,1])
#        inv_var_plot([best_fit], [best_id], 'slink_scree_cluster_' + str(len(c)))

def cluster_scree_plots ():
    points = import_csv(find_files('RNA_data.csv')[0])['PDB-Data'][:,:7]
    print(points.shape)
    max_var = 13745.3721915
    clusters = import_lists(find_files('RNA_data*multiSLINK_result.csv')[0])
    clusters = [np.array(sorted(x)) for x in clusters[:-1]]
    scree_data = []
    scree_data_euclid = []
    scree_labels = []
    euclid_labels = []
    clusters = list(reversed(sorted(clusters, key=len)))
    for i,c in enumerate(clusters):
        if not len(c) in [226, 133, 60, 52]:
            continue
        p = points[c]
        val, vec = la.eig(np.cov(euclideanize(p.copy(), mode='mean').T))
        val = np.hstack((np.zeros(1), np.cumsum(np.sort(val)[::-1])))
        val = np.max(val) - val
        if val[1] < 0.1 * max_var:
            continue
        name = ('pre-cluster 5 ' if len(c) == 226 else
                ('pre-cluster 8 ' if len(c) == 133 else
                 ('pre-cluster 11 ' if len(c) == 60 else 'pre-cluster 12 ')))
        print(p.shape)
        best_fit = None; suffix = None
        for [invert, mode] in [[False, 'gap'], [True, 'gap'],
                               [False, 'mean'], [True, 'mean']]:
            print(('SO ' if invert else 'SI ') + mode)
            sphere_points, means, half = RESHify_1D(p, invert, mode)
            spheres, projected_points, distances = pns_loop(sphere_points, degenerate=half, verbose=False, mode='scale')
            for s in spheres[:-1]: print(s.radius)
            sys.stdout.flush()
            center_point = unRESHify_1D(unfold_points(as_matrix(projected_points[-1]), spheres[:-1]), means, half)
            unfolded_angles = [unRESHify_1D(unfold_points(pts, spheres[:i+1]), means, half) for i, pts in enumerate(projected_points[:-1])]
            var_data = []
            for ang in unfolded_angles:
                var_data.append(np.mean(torus_distances(ang, p)**2))
            var_data.append(np.mean(torus_distances(center_point, p)**2))
            var_data = np.array(var_data)[::-1]
            print(var_data)
            sys.stdout.flush()
            if (best_fit is None) or (var_data[1] < best_fit[1]):
                print('better fit!')
                best_fit = var_data
                suffix = ('SO ' if invert else 'SI ') + ('GC' if mode=='gap' else 'MC')
        if val[1] > max(0.2 * max_var, best_fit[1] * 1.5):
            scree_data.append(np.array([[i, 100 * v / max_var] for i,v in enumerate(best_fit)]))
            scree_data_euclid.append(np.array([[i, 100 * v / max_var] for i,v in enumerate(val[:-1])]))
            scree_labels.append((name+suffix).replace('_',' '))
            euclid_labels.append((name[:-1]).replace('_',' '))
    inv_var_plot(scree_data, scree_labels, 'full_data_slink_scree')
    inv_var_plot(scree_data_euclid, euclid_labels, 'full_data_slink_scree_euclidean')


if __name__ == '__main__':
#    new_multi_slink(main_all())
#    geoPCA_orig_data()
#    scree()
#    plots_2d_vs_7d()
#    sec_struct_for_3()
#    paper_plots()
#    t2_testing()
#    check_files()
#    plot_residues()
#    cluster_sizes()
#    make_8()
#    check_chains()
#    cluster_overview()
    cluster_scree_plots()
